import numpy as np

from math import sqrt, sin, cos, acos, atan2, trunc, pi
from .oc_geometry import nekmin, wd_mesh_fill, wd_recompute_neck
import libphoebe
import os
from phoebe.distortions.roche import *
# from scipy.spatial import KDTree

import logging
logger = logging.getLogger("MESH_WD")
logger.addHandler(logging.NullHandler())


class MeshVertex:
    def __init__(self, r, dpdx, dpdy, dpdz, *args):
        """
        """
        # Normalized normal:
        #~ n = np.array([dpdx(r, *args), dpdy(r, *args), dpdz(r, *args)])
        #~ n /= np.sqrt(np.sum(n*n))
        nx = dpdx(r, *args)
        ny = dpdy(r, *args)
        nz = dpdz(r, *args)
        nn = sqrt(nx*nx+ny*ny+nz*nz)
        nx /= nn
        ny /= nn
        nz /= nn

        # Now we choose the first tangential direction. We have a whole
        # plane to choose from, so we'll choose something convenient.
        # The choice here is to set one tangential coordinate to 0. From
        # n \dot t=0: nx tx + ny ty = 0 => tx = -ny, ty = nx.
        # Since we are normalizing, we need to be careful that we don't
        # divide by a small number, hence the two prescriptions, either
        # for tz = 0 or for ty = 0.

        if nx > 0.5 or ny > 0.5:
            nn = sqrt(ny*ny+nx*nx)
            t1x = ny/nn
            t1y = -nx/nn
            t1z = 0.0
        else:
            nn = sqrt(nx*nx+nz*nz)
            t1x = -nz/nn
            t1y = 0.0
            t1z = nx/nn

        t2x = ny*t1z - nz*t1y
        t2y = nz*t1x - nx*t1z
        t2z = nx*t1y - ny*t1x

        self.r = r
        self.n = np.array((nx, ny, nz))
        self.t1 = np.array((t1x, t1y, t1z))
        self.t2 = np.array((t2x, t2y, t2z))

    def __repr__(self):
        repstr  = " r = (% 3.3f, % 3.3f, % 3.3f)\t" % (self.r[0],  self.r[1], self.r[2])
        repstr += " n = (% 3.3f, % 3.3f, % 3.3f)\t" % (self.n[0],  self.n[1], self.n[2])
        repstr += "t1 = (% 3.3f, % 3.3f, % 3.3f)\t" % (self.t1[0], self.t1[1], self.t1[2])
        repstr += "t2 = (% 3.3f, % 3.3f, % 3.3f)"   % (self.t2[0], self.t2[1], self.t2[2])
        return repstr

def project_onto_potential(r, pot_name, *args):
    """
    TODO: add documentation
    """

    pot = globals()[pot_name]
    dpdx = globals()['d%sdx'%(pot_name)]
    dpdy = globals()['d%sdy'%(pot_name)]
    dpdz = globals()['d%sdz'%(pot_name)]
    dpdr = globals()['d%sdr'%(pot_name)]

    n_iter = 0

    rmag, rmag0 = np.sqrt((r*r).sum()), 0
    lam, nu = r[0]/rmag, r[2]/rmag
    dc = np.array((lam, np.sqrt(1-lam*lam-nu*nu), nu)) # direction cosines -- must not change during reprojection
    D, q, F, p0 = args

    while np.abs(rmag-rmag0) > 1e-12 and n_iter < 100:
        rmag0 = rmag
        rmag = rmag0 - pot(rmag0*dc, *args)/dpdr(rmag0*dc, *args[:-1])
        n_iter += 1
    if n_iter == 100:
        logger.warning('projection did not converge')

    r = rmag*dc

    return MeshVertex(r, dpdx, dpdy, dpdz, *args[:-1])

def discretize_wd_style(N, q, F, d, Phi):
    """
    TODO: add documentation

    New implementation. I'll make this work first, then document.
    """

    DEBUG = False

    Ts = []

    potential = 'BinaryRoche'
    r0 = libphoebe.roche_pole(q, F, d, Phi)

    # The following is a hack that needs to go!
    pot_name = potential
    dpdx = globals()['d%sdx'%(pot_name)]
    dpdy = globals()['d%sdy'%(pot_name)]
    dpdz = globals()['d%sdz'%(pot_name)]

    if DEBUG:
        import matplotlib.pyplot as plt
        from matplotlib.path import Path
        import matplotlib.patches as patches

        fig = plt.figure()
        ax1 = fig.add_subplot(131)
        ax2 = fig.add_subplot(132)
        ax3 = fig.add_subplot(133)
        ax1.set_xlim(-0.3, 0.3) # -1.6 1.6
        ax1.set_ylim(-0.3, 0.3)
        ax2.set_xlim(-0.3, 0.3)
        ax2.set_ylim(-0.3, 0.3)
        ax3.set_xlim(-0.3, 0.3)
        ax3.set_ylim(-0.3, 0.3)
        ax1.set_xlabel('x')
        ax1.set_ylabel('y')
        ax2.set_xlabel('x')
        ax2.set_ylabel('z')
        ax3.set_xlabel('y')
        ax3.set_ylabel('z')

    # Rectangle centers:
    theta = np.array([np.pi/2*(k-0.5)/N for k in range(1, N+2)])
    phi = np.array([[np.pi*(l-0.5)/Mk for l in range(1, Mk+1)] for Mk in np.array(1 + 1.3*N*np.sin(theta), dtype=int)])

    for t in range(len(theta)-1):
        dtheta = theta[t+1]-theta[t]
        for i in range(len(phi[t])):
            dphi = phi[t][1]-phi[t][0]

            # Project the vertex onto the potential; this will be our center point:
            rc = np.array((r0*sin(theta[t])*cos(phi[t][i]), r0*sin(theta[t])*sin(phi[t][i]), r0*cos(theta[t])))
            vc = project_onto_potential(rc, potential, d, q, F, Phi).r

            # Next we need to find the tangential plane, which we'll get by finding the normal,
            # which is the negative of the gradient:
            nc = np.array((-dpdx(vc, d, q, F), -dpdy(vc, d, q, F), -dpdz(vc, d, q, F)))

            # Then we need to find the intercontext of +/-dtheta/dphi-deflected
            # radius vectors with the tangential plane. We do that by solving
            #
            #   d = [(p0 - l0) \dot n] / (l \dot n),
            #
            # where p0 and l0 are reference points on the plane and on the line,
            # respectively, n is the normal vector, and l is the line direction
            # vector. For convenience l0 can be set to 0, and p0 is just vc. d
            # then measures the distance from the origin along l.

            l1 = np.array((sin(theta[t]-dtheta/2)*cos(phi[t][i]-dphi/2), sin(theta[t]-dtheta/2)*sin(phi[t][i]-dphi/2), cos(theta[t]-dtheta/2)))
            l2 = np.array((sin(theta[t]-dtheta/2)*cos(phi[t][i]+dphi/2), sin(theta[t]-dtheta/2)*sin(phi[t][i]+dphi/2), cos(theta[t]-dtheta/2)))
            l3 = np.array((sin(theta[t]+dtheta/2)*cos(phi[t][i]+dphi/2), sin(theta[t]+dtheta/2)*sin(phi[t][i]+dphi/2), cos(theta[t]+dtheta/2)))
            l4 = np.array((sin(theta[t]+dtheta/2)*cos(phi[t][i]-dphi/2), sin(theta[t]+dtheta/2)*sin(phi[t][i]-dphi/2), cos(theta[t]+dtheta/2)))

            r1 = np.dot(vc, nc) / np.dot(l1, nc) * l1
            r2 = np.dot(vc, nc) / np.dot(l2, nc) * l2
            r3 = np.dot(vc, nc) / np.dot(l3, nc) * l3
            r4 = np.dot(vc, nc) / np.dot(l4, nc) * l4

            # This sorts out the vertices, now we need to fudge the surface
            # area. WD does not take curvature of the equipotential at vc
            # into account, so the surface area computed from these vertex-
            # delimited surfaces will generally be different from what WD
            # computes. Thus, we compute the surface area the same way WD
            # does it and assign it to each element even though that isn't
            # quite its area:
            #
            #   dsigma = || r^2 sin(theta)/cos(gamma) dtheta dphi ||,
            #
            # where gamma is the angle between l and n.

            cosgamma = np.dot(vc, nc)/np.sqrt(np.dot(vc, vc))/np.sqrt(np.dot(nc, nc))
            dsigma = np.abs(np.dot(vc, vc)*np.sin(theta[t])/cosgamma*dtheta*dphi)

            # Temporary addition: triangle areas: ######################
            side1 = sqrt((r1[0]-r2[0])**2 + (r1[1]-r2[1])**2 + (r1[2]-r2[2])**2)
            side2 = sqrt((r1[0]-r3[0])**2 + (r1[1]-r3[1])**2 + (r1[2]-r3[2])**2)
            side3 = sqrt((r2[0]-r3[0])**2 + (r2[1]-r3[1])**2 + (r2[2]-r3[2])**2)
            s = 0.5*(side1 + side2 + side3)

            dsigma_t_sq = s*(s-side1)*(s-side2)*(s-side3)
            dsigma_t = sqrt(dsigma_t_sq) if dsigma_t_sq > 0 else 0.0
            ############################################################

            if DEBUG:
                fc = 'orange'

                verts = [(r1[0], r1[1]), (r2[0], r2[1]), (r3[0], r3[1]), (r4[0], r4[1]), (r1[0], r1[1])]
                codes = [Path.MOVETO, Path.LINETO, Path.LINETO, Path.LINETO, Path.CLOSEPOLY]
                path = Path(verts, codes)
                patch = patches.PathPatch(path, facecolor=fc, lw=2)
                ax1.add_patch(patch)

                verts = [(r1[0], r1[2]), (r2[0], r2[2]), (r3[0], r3[2]), (r4[0], r4[2]), (r1[0], r1[2])]
                codes = [Path.MOVETO, Path.LINETO, Path.LINETO, Path.LINETO, Path.CLOSEPOLY]
                path = Path(verts, codes)
                patch = patches.PathPatch(path, facecolor=fc, lw=2)
                ax2.add_patch(patch)

                verts = [(r1[1], r1[2]), (r2[1], r2[2]), (r3[1], r3[2]), (r4[1], r4[2]), (r1[1], r1[2])]
                codes = [Path.MOVETO, Path.LINETO, Path.LINETO, Path.LINETO, Path.CLOSEPOLY]
                path = Path(verts, codes)
                patch = patches.PathPatch(path, facecolor=fc, lw=2)
                ax3.add_patch(patch)

            # Ts.append(np.array((vc[0], vc[1], vc[2], dsigma/2, r1[0], r1[1], r1[2], r2[0], r2[1], r2[2], r3[0], r3[1], r3[2], nc[0], nc[1], nc[2])))
            # Ts.append(np.array((vc[0], vc[1], vc[2], dsigma/2, r3[0], r3[1], r3[2], r4[0], r4[1], r4[2], r1[0], r1[1], r1[2], nc[0], nc[1], nc[2])))

            # # Instead of recomputing all quantities, just reflect over the y- and z-directions.
            # Ts.append(np.array((vc[0], -vc[1],  vc[2], dsigma/2, r1[0], -r1[1],  r1[2], r2[0], -r2[1],  r2[2], r3[0], -r3[1],  r3[2], nc[0], -nc[1], nc[2])))
            # Ts.append(np.array((vc[0], -vc[1],  vc[2], dsigma/2, r3[0], -r3[1],  r3[2], r4[0], -r4[1],  r4[2], r1[0], -r1[1],  r1[2], nc[0], -nc[1], nc[2])))

            # Ts.append(np.array((vc[0],  vc[1], -vc[2], dsigma/2, r1[0],  r1[1], -r1[2], r2[0],  r2[1], -r2[2], r3[0],  r3[1], -r3[2], nc[0],  nc[1], -nc[2])))
            # Ts.append(np.array((vc[0],  vc[1], -vc[2], dsigma/2, r3[0],  r3[1], -r3[2], r4[0],  r4[1], -r4[2], r1[0],  r1[1], -r1[2], nc[0],  nc[1], -nc[2])))

            # Ts.append(np.array((vc[0], -vc[1], -vc[2], dsigma/2, r1[0], -r1[1], -r1[2], r2[0], -r2[1], -r2[2], r3[0], -r3[1], -r3[2], nc[0], -nc[1], -nc[2])))
            # Ts.append(np.array((vc[0], -vc[1], -vc[2], dsigma/2, r3[0], -r3[1], -r3[2], r4[0], -r4[1], -r4[2], r1[0], -r1[1], -r1[2], nc[0], -nc[1], -nc[2])))

            # FOR TESTING - report theta/phi for each triangle
            # uncomment the above original version eventually
            Ts.append(np.array((vc[0], vc[1], vc[2], dsigma/2, r1[0], r1[1], r1[2], r2[0], r2[1], r2[2], r3[0], r3[1], r3[2], nc[0], nc[1], nc[2], theta[t], phi[t][0], dsigma_t)))
            Ts.append(np.array((vc[0], vc[1], vc[2], dsigma/2, r3[0], r3[1], r3[2], r4[0], r4[1], r4[2], r1[0], r1[1], r1[2], nc[0], nc[1], nc[2], theta[t], phi[t][0], dsigma_t)))

            # Instead of recomputing all quantities, just reflect over the y- and z-directions.
            Ts.append(np.array((vc[0], -vc[1],  vc[2], dsigma/2, r1[0], -r1[1],  r1[2], r2[0], -r2[1],  r2[2], r3[0], -r3[1],  r3[2], nc[0], -nc[1], nc[2], theta[t], -phi[t][0], dsigma_t)))
            Ts.append(np.array((vc[0], -vc[1],  vc[2], dsigma/2, r3[0], -r3[1],  r3[2], r4[0], -r4[1],  r4[2], r1[0], -r1[1],  r1[2], nc[0], -nc[1], nc[2], theta[t], -phi[t][0], dsigma_t)))

            Ts.append(np.array((vc[0],  vc[1], -vc[2], dsigma/2, r1[0],  r1[1], -r1[2], r2[0],  r2[1], -r2[2], r3[0],  r3[1], -r3[2], nc[0],  nc[1], -nc[2], np.pi-theta[t], phi[t][0], dsigma_t)))
            Ts.append(np.array((vc[0],  vc[1], -vc[2], dsigma/2, r3[0],  r3[1], -r3[2], r4[0],  r4[1], -r4[2], r1[0],  r1[1], -r1[2], nc[0],  nc[1], -nc[2], np.pi-theta[t], phi[t][0], dsigma_t)))

            Ts.append(np.array((vc[0], -vc[1], -vc[2], dsigma/2, r1[0], -r1[1], -r1[2], r2[0], -r2[1], -r2[2], r3[0], -r3[1], -r3[2], nc[0], -nc[1], -nc[2], np.pi-theta[t], -phi[t][0], dsigma_t)))
            Ts.append(np.array((vc[0], -vc[1], -vc[2], dsigma/2, r3[0], -r3[1], -r3[2], r4[0], -r4[1], -r4[2], r1[0], -r1[1], -r1[2], nc[0], -nc[1], -nc[2], np.pi-theta[t], -phi[t][0], dsigma_t)))

    if DEBUG:
        plt.show()

    # Assemble a mesh table:
    table = np.array(Ts)
    return table

def discretize_wd_style_oc(N, q, F, d, Phi,recompute_neck=True):

    Ts = []
    Tstart = [0,]

    # for testing of where things start to diverge:
    breaks0, breaks1 = [], []
    vcs0, vcs1 = [], []

    potential = 'BinaryRoche'
    q_1, Phi_1 = q, Phi
    q_2, Phi_2 = 1./q, Phi/q + 0.5*(q-1)/q

    xmin1, z1 = nekmin(Phi_1,q_1,0.5,0.05,0.05)
    # xmin1 = (xminz1+xminy1)/2.

    xmin2 = d - xmin1
    #xminz2, xminy2, y2, z2 = nekmin(Phi_2,q_2,0.5,0.05,0.05)

    # first, find closest neck points to mark them for fractional area computation
    for obj, q, Phi, xmin in zip([0,1],[q_1,q_2],[Phi_1,Phi_2],[xmin1,xmin2]):

        r0 = libphoebe.roche_pole(q, F, d, Phi)
        # The following is a hack that needs to go!
        pot_name = potential
        dpdx = globals()['d%sdx'%(pot_name)]
        dpdy = globals()['d%sdy'%(pot_name)]
        dpdz = globals()['d%sdz'%(pot_name)]

        # Rectangle centers:
        theta = np.array([np.pi/2*(k-0.5)/N for k in range(1, N+2)])
        phi = np.array([[np.pi*(l-0.5)/Mk for l in range(1, Mk+1)] for Mk in np.array(1 + 1.3*N*np.sin(theta), dtype=int)])

        for t in range(len(theta)-1):
            dtheta = theta[t+1]-theta[t]
            for i in range(len(phi[t])):
                dphi = phi[t][1]-phi[t][0]

                # Project the vertex onto the potential; this will be our center point:
                rc = np.array((r0*sin(theta[t])*cos(phi[t][i]), r0*sin(theta[t])*sin(phi[t][i]), r0*cos(theta[t])))
                vc = project_onto_potential(rc, potential, d, q, F, Phi).r

                if (vc[0] < 0. and vc[0] > -1.) or abs(vc[0]) <= xmin:
                    # do this for separate components cause they may have different diverging strips
                    if obj==0 and abs(vc[1])>=1e-16:
                        vcs0.append(np.array([vc[0],vc[1],vc[2],theta[t]]))
                    elif obj==1 and abs(vc[1])>=1e-16:
                        vcs1.append(np.array([vc[0],vc[1],vc[2],theta[t]]))
                else:
                    if obj==0 and abs(vc[1])>=1e-16:
                        breaks0.append(theta[t])
                    elif obj==1 and abs(vc[1])>=1e-16:
                        breaks1.append(theta[t])

    vcs0,vcs1 = np.array(vcs0), np.array(vcs1)
    #breaks0, breaks1 = np.array(breaks0), np.array(breaks1)
    thetas0, thetas1 = set(breaks0), set(breaks1)

    vcs_breaks0 = []
    vcs_breaks1 = []
    # go through strips of thetas where things diverge and mark last created center

    for theta in thetas0:
        vcs_strip = vcs0[vcs0[:,-1]==theta]
        vcs_breaks0.append([vcs_strip[0][0],vcs_strip[0][1],vcs_strip[0][2]])

    for theta in thetas1:
        vcs_strip = vcs1[vcs1[:,-1]==theta]
        vcs_breaks1.append([vcs_strip[0][0],vcs_strip[0][1],vcs_strip[0][2]])

    vcs_breaks0 = np.array(vcs_breaks0)
    vcs_breaks1 = np.array(vcs_breaks1)

    for obj, q, Phi, xmin, vcs_breaks in zip([0,1],[q_1,q_2],[Phi_1,Phi_2],[xmin1,xmin2],[vcs_breaks0,vcs_breaks1]):

        r0 = libphoebe.roche_pole(q, F, d, Phi)
        # The following is a hack that needs to go!
        pot_name = potential
        dpdx = globals()['d%sdx'%(pot_name)]
        dpdy = globals()['d%sdy'%(pot_name)]
        dpdz = globals()['d%sdz'%(pot_name)]

        # Rectangle centers:
        theta = np.array([np.pi/2*(k-0.5)/N for k in range(1, N+2)])
        phi = np.array([[np.pi*(l-0.5)/Mk for l in range(1, Mk+1)] for Mk in np.array(1 + 1.3*N*np.sin(theta), dtype=int)])


        for t in range(len(theta)-1):
            dtheta = theta[t+1]-theta[t]
            for i in range(len(phi[t])):
                dphi = phi[t][1]-phi[t][0]

                # Project the vertex onto the potential; this will be our center point:
                # print "projecting center"
                rc = np.array((r0*sin(theta[t])*cos(phi[t][i]), r0*sin(theta[t])*sin(phi[t][i]), r0*cos(theta[t])))
                vc = project_onto_potential(rc, potential, d, q, F, Phi).r

                if ((vc[0] < 0. and vc[0] > -1.) or abs(vc[0]) <= xmin) and abs(vc[1])>=1e-16:

                    # Next we need to find the tangential plane, which we'll get by finding the normal,
                    # which is the negative of the gradient:
                    nc = np.array((-dpdx(vc, d, q, F), -dpdy(vc, d, q, F), -dpdz(vc, d, q, F)))

                    # Then we need to find the intercontext of +/-dtheta/dphi-deflected
                    # radius vectors with the tangential plane. We do that by solving
                    #
                    #   d = [(p0 - l0) \dot n] / (l \dot n),
                    #
                    # where p0 and l0 are reference points on the plane and on the line,
                    # respectively, n is the normal vector, and l is the line direction
                    # vector. For convenience l0 can be set to 0, and p0 is just vc. d
                    # then measures the distance from the origin along l.

                    l1 = np.array((sin(theta[t]-dtheta/2)*cos(phi[t][i]-dphi/2), sin(theta[t]-dtheta/2)*sin(phi[t][i]-dphi/2), cos(theta[t]-dtheta/2)))
                    l2 = np.array((sin(theta[t]-dtheta/2)*cos(phi[t][i]+dphi/2), sin(theta[t]-dtheta/2)*sin(phi[t][i]+dphi/2), cos(theta[t]-dtheta/2)))
                    l3 = np.array((sin(theta[t]+dtheta/2)*cos(phi[t][i]+dphi/2), sin(theta[t]+dtheta/2)*sin(phi[t][i]+dphi/2), cos(theta[t]+dtheta/2)))
                    l4 = np.array((sin(theta[t]+dtheta/2)*cos(phi[t][i]-dphi/2), sin(theta[t]+dtheta/2)*sin(phi[t][i]-dphi/2), cos(theta[t]+dtheta/2)))

                    r1 = np.dot(vc, nc) / np.dot(l1, nc) * l1
                    r2 = np.dot(vc, nc) / np.dot(l2, nc) * l2
                    r3 = np.dot(vc, nc) / np.dot(l3, nc) * l3
                    r4 = np.dot(vc, nc) / np.dot(l4, nc) * l4

                    # This sorts out the vertices, now we need to fudge the surface
                    # area. WD does not take curvature of the equipotential at vc
                    # into account, so the surface area computed from these vertex-
                    # delimited surfaces will generally be different from what WD
                    # computes. Thus, we compute the surface area the same way WD
                    # does it and assign it to each element even though that isn't
                    # quite its area:
                    #
                    #   dsigma = || r^2 sin(theta)/cos(gamma) dtheta dphi ||,
                    #
                    # where gamma is the angle between l and n.

                    cosgamma = np.dot(vc, nc)/np.sqrt(np.dot(vc, vc))/np.sqrt(np.dot(nc, nc))
                    dsigma = np.abs(np.dot(vc, vc)*np.sin(theta[t])/cosgamma*dtheta*dphi)

                    # Temporary addition: triangle areas: ######################
                    side1 = sqrt((r1[0]-r2[0])**2 + (r1[1]-r2[1])**2 + (r1[2]-r2[2])**2)
                    side2 = sqrt((r1[0]-r3[0])**2 + (r1[1]-r3[1])**2 + (r1[2]-r3[2])**2)
                    side3 = sqrt((r2[0]-r3[0])**2 + (r2[1]-r3[1])**2 + (r2[2]-r3[2])**2)
                    s = 0.5*(side1 + side2 + side3)

                    dsigma_t_sq = s*(s-side1)*(s-side2)*(s-side3)
                    dsigma_t = sqrt(dsigma_t_sq) if dsigma_t_sq > 0 else 0.0
                    ############################################################

                    #if abs(r1[0]) <= xmin and abs(r2[0]) <= xmin and abs(r3[0]) <= xmin and abs(r4[0]) <= xmin:

                    # check whether the trapezoids cross the neck and if so put them back into their place
                    # do the same for end traezoids also

                    if (np.array([abs(r1[0]),abs(r2[0]),abs(r3[0]),abs(r4[0])])>xmin).any() or vc in vcs_breaks:
                        r1,r2,r3,r4,dsigma_t_new = wd_mesh_fill(xmin,y1,z1,r1,r2,r3,r4)
                        if dsigma_t != 0:
                            frac_area = dsigma_t_new/dsigma_t
                            dsigma = dsigma*frac_area
                            dsigma_t = dsigma_t*frac_area
                        else:
                            dsigma_t = dsigma_t_new
                            dsigma = 2*dsigma_t_new

                        if recompute_neck:
                            #vc,nc,r1,r2,r3,r4,dsigma,dsigma_t,phi_half,dphi_new=wd_recompute_neck(r0,xmin,y1,z1,r1,r2,r3,r4,theta[t],dtheta,potential,d,q,F,Phi)
                            disgma = wd_recompute_neck(r0,xmin,y1,z1,r1,r2,r3,r4,theta[t],dtheta,potential,d,q,F,Phi)
                            #phi[t][i],dphi=phi_half,dphi_new

                        # If all of the vertices are on one side of the minimum, keep trapezoids as they are

                        # Ts.append(np.array((vc[0], vc[1], vc[2], dsigma/2, r1[0], r1[1], r1[2], r2[0], r2[1], r2[2], r3[0], r3[1], r3[2], nc[0], nc[1], nc[2])))
                        # Ts.append(np.array((vc[0], vc[1], vc[2], dsigma/2, r3[0], r3[1], r3[2], r4[0], r4[1], r4[2], r1[0], r1[1], r1[2], nc[0], nc[1], nc[2])))

                        # # Instead of recomputing all quantities, just reflect over the y- and z-directions.
                        # Ts.append(np.array((vc[0], -vc[1],  vc[2], dsigma/2, r1[0], -r1[1],  r1[2], r2[0], -r2[1],  r2[2], r3[0], -r3[1],  r3[2], nc[0], -nc[1], nc[2])))
                        # Ts.append(np.array((vc[0], -vc[1],  vc[2], dsigma/2, r3[0], -r3[1],  r3[2], r4[0], -r4[1],  r4[2], r1[0], -r1[1],  r1[2], nc[0], -nc[1], nc[2])))

                        # Ts.append(np.array((vc[0],  vc[1], -vc[2], dsigma/2, r1[0],  r1[1], -r1[2], r2[0],  r2[1], -r2[2], r3[0],  r3[1], -r3[2], nc[0],  nc[1], -nc[2])))
                        # Ts.append(np.array((vc[0],  vc[1], -vc[2], dsigma/2, r3[0],  r3[1], -r3[2], r4[0],  r4[1], -r4[2], r1[0],  r1[1], -r1[2], nc[0],  nc[1], -nc[2])))

                        # Ts.append(np.array((vc[0], -vc[1], -vc[2], dsigma/2, r1[0], -r1[1], -r1[2], r2[0], -r2[1], -r2[2], r3[0], -r3[1], -r3[2], nc[0], -nc[1], -nc[2])))
                        # Ts.append(np.array((vc[0], -vc[1], -vc[2], dsigma/2, r3[0], -r3[1], -r3[2], r4[0], -r4[1], -r4[2], r1[0], -r1[1], -r1[2], nc[0], -nc[1], -nc[2])))

                        # FOR TESTING - report theta/phi for each triangle
                        # uncomment the above original version eventually


                    # Temporary addition: vertex normals: ######################
                    # print "projecting vertices "
                    v1 = project_onto_potential(r1, potential, d, q, F, Phi).r
                    v2 = project_onto_potential(r2, potential, d, q, F, Phi).r
                    v3 = project_onto_potential(r3, potential, d, q, F, Phi).r
                    v4 = project_onto_potential(r4, potential, d, q, F, Phi).r

                    n1 = np.array((-dpdx(v1, d, q, F), -dpdy(v1, d, q, F), -dpdz(v1, d, q, F)))
                    n2 = np.array((-dpdx(v2, d, q, F), -dpdy(v2, d, q, F), -dpdz(v2, d, q, F)))
                    n3 = np.array((-dpdx(v3, d, q, F), -dpdy(v3, d, q, F), -dpdz(v3, d, q, F)))
                    n4 = np.array((-dpdx(v4, d, q, F), -dpdy(v4, d, q, F), -dpdz(v4, d, q, F)))

                    ############################################################

                    # Ts.append(np.array((vc[0], vc[1], vc[2], dsigma/2, r1[0], r1[1], r1[2], r2[0], r2[1], r2[2], r3[0], r3[1], r3[2], nc[0], nc[1], nc[2])))
                    # Ts.append(np.array((vc[0], vc[1], vc[2], dsigma/2, r3[0], r3[1], r3[2], r4[0], r4[1], r4[2], r1[0], r1[1], r1[2], nc[0], nc[1], nc[2])))

                    # # Instead of recomputing all quantities, just reflect over the y- and z-directions.
                    # Ts.append(np.array((vc[0], -vc[1],  vc[2], dsigma/2, r1[0], -r1[1],  r1[2], r2[0], -r2[1],  r2[2], r3[0], -r3[1],  r3[2], nc[0], -nc[1], nc[2])))
                    # Ts.append(np.array((vc[0], -vc[1],  vc[2], dsigma/2, r3[0], -r3[1],  r3[2], r4[0], -r4[1],  r4[2], r1[0], -r1[1],  r1[2], nc[0], -nc[1], nc[2])))

                    # Ts.append(np.array((vc[0],  vc[1], -vc[2], dsigma/2, r1[0],  r1[1], -r1[2], r2[0],  r2[1], -r2[2], r3[0],  r3[1], -r3[2], nc[0],  nc[1], -nc[2])))
                    # Ts.append(np.array((vc[0],  vc[1], -vc[2], dsigma/2, r3[0],  r3[1], -r3[2], r4[0],  r4[1], -r4[2], r1[0],  r1[1], -r1[2], nc[0],  nc[1], -nc[2])))

                    # Ts.append(np.array((vc[0], -vc[1], -vc[2], dsigma/2, r1[0], -r1[1], -r1[2], r2[0], -r2[1], -r2[2], r3[0], -r3[1], -r3[2], nc[0], -nc[1], -nc[2])))
                    # Ts.append(np.array((vc[0], -vc[1], -vc[2], dsigma/2, r3[0], -r3[1], -r3[2], r4[0], -r4[1], -r4[2], r1[0], -r1[1], -r1[2], nc[0], -nc[1], -nc[2])))

                    # FOR TESTING - report theta/phi for each triangle
                    # uncomment the above original version eventually
                    if obj==0:
                        Ts.append(np.array((vc[0], vc[1], vc[2], dsigma/2, r3[0], r3[1], r3[2], r4[0], r4[1], r4[2], r1[0], r1[1], r1[2], nc[0], nc[1], nc[2], theta[t], phi[t][0], dsigma_t, n3[0], n3[1], n3[2], n4[0], n4[1], n4[2], n1[0], n1[1], n1[2])))
                        Ts.append(np.array((vc[0], vc[1], vc[2], dsigma/2, r1[0], r1[1], r1[2], r2[0], r2[1], r2[2], r3[0], r3[1], r3[2], nc[0], nc[1], nc[2], theta[t], phi[t][0], dsigma_t, n1[0], n1[1], n1[2], n2[0], n2[1], n2[2], n3[0], n3[1], n3[2])))

                    else:
                        Ts.append(np.array((-vc[0]+d, vc[1], vc[2], dsigma/2, -r3[0]+d, r3[1], r3[2], -r4[0]+d, r4[1], r4[2], -r1[0]+d, r1[1], r1[2], -nc[0], nc[1], nc[2], theta[t], phi[t][0], dsigma_t, -n3[0], n3[1], n3[2], -n4[0], n4[1], n4[2], -n1[0], n1[1], n1[2])))
                        Ts.append(np.array((-vc[0]+d, vc[1], vc[2], dsigma/2, -r1[0]+d, r1[1], r1[2], -r2[0]+d, r2[1], r2[2], -r3[0]+d, r3[1], r3[2], -nc[0], nc[1], nc[2], theta[t], phi[t][0], dsigma_t, -n1[0], n1[1], n1[2], -n2[0], n2[1], n2[2], -n3[0], n3[1], n3[2])))

        for T in reversed(Ts[Tstart[-1]:]):
            Ts.append(np.array((T[0], -T[1],  T[2], T[3], T[4], -T[5],  T[6], T[7], -T[8],  T[9], T[10], -T[11],  T[12], T[13], -T[14], T[15], T[16], -T[17], T[18], T[19], -T[20],  T[21], T[22], -T[23],  T[24], T[25], -T[26],  T[27])))
        Tstart.append(len(Ts))

    for i in range(len(Tstart)-1):
        for T in Ts[Tstart[-2]:Tstart[-1]]:
            Ts.append(np.array((T[0],  T[1], -T[2], T[3], T[4],  T[5], -T[6], T[7],  T[8], -T[9], T[10],  T[11], -T[12], T[13],  T[14], -T[15], np.pi-T[16], T[17], T[18], T[19],  T[20], -T[21], T[22],  T[23], -T[24], T[25],  T[26], -T[27])))
        Tstart.pop(-1)

                    # if obj == 0:
                    #     Ts.append(np.array((vc[0], vc[1], vc[2], dsigma/2, r1[0], r1[1], r1[2], r2[0], r2[1], r2[2], r3[0], r3[1], r3[2], nc[0], nc[1], nc[2], theta[t], phi[t][0], dsigma_t)))
                    #     Ts.append(np.array((vc[0], vc[1], vc[2], dsigma/2, r3[0], r3[1], r3[2], r4[0], r4[1], r4[2], r1[0], r1[1], r1[2], nc[0], nc[1], nc[2], theta[t], phi[t][0], dsigma_t)))

                    #     # Instead of recomputing all quantities, just reflect over the y- and z-directions.
                    #     Ts.append(np.array((vc[0], -vc[1],  vc[2], dsigma/2, r1[0], -r1[1],  r1[2], r2[0], -r2[1],  r2[2], r3[0], -r3[1],  r3[2], nc[0], -nc[1], nc[2], theta[t], -phi[t][0], dsigma_t)))
                    #     Ts.append(np.array((vc[0], -vc[1],  vc[2], dsigma/2, r3[0], -r3[1],  r3[2], r4[0], -r4[1],  r4[2], r1[0], -r1[1],  r1[2], nc[0], -nc[1], nc[2], theta[t], -phi[t][0], dsigma_t)))

                    #     Ts.append(np.array((vc[0],  vc[1], -vc[2], dsigma/2, r1[0],  r1[1], -r1[2], r2[0],  r2[1], -r2[2], r3[0],  r3[1], -r3[2], nc[0],  nc[1], -nc[2], np.pi-theta[t], phi[t][0], dsigma_t)))
                    #     Ts.append(np.array((vc[0],  vc[1], -vc[2], dsigma/2, r3[0],  r3[1], -r3[2], r4[0],  r4[1], -r4[2], r1[0],  r1[1], -r1[2], nc[0],  nc[1], -nc[2], np.pi-theta[t], phi[t][0], dsigma_t)))

                    #     Ts.append(np.array((vc[0], -vc[1], -vc[2], dsigma/2, r1[0], -r1[1], -r1[2], r2[0], -r2[1], -r2[2], r3[0], -r3[1], -r3[2], nc[0], -nc[1], -nc[2], np.pi-theta[t], -phi[t][0], dsigma_t)))
                    #     Ts.append(np.array((vc[0], -vc[1], -vc[2], dsigma/2, r3[0], -r3[1], -r3[2], r4[0], -r4[1], -r4[2], r1[0], -r1[1], -r1[2], nc[0], -nc[1], -nc[2], np.pi-theta[t], -phi[t][0], dsigma_t)))
                    # else:
                    #     Ts.append(np.array((-vc[0]+d, vc[1], vc[2], dsigma/2, -r1[0]+d, r1[1], r1[2], -r2[0]+d, r2[1], r2[2], -r3[0]+d, r3[1], r3[2], -nc[0], nc[1], nc[2], theta[t], phi[t][0], dsigma_t)))
                    #     Ts.append(np.array((-vc[0]+d, vc[1], vc[2], dsigma/2, -r3[0]+d, r3[1], r3[2], -r4[0]+d, r4[1], r4[2], -r1[0]+d, r1[1], r1[2], -nc[0], nc[1], nc[2], theta[t], phi[t][0], dsigma_t)))

                    #     # Instead of recomputing all quantities, just reflect over the y- and z-directions.
                    #     Ts.append(np.array((-vc[0]+d, -vc[1],  vc[2], dsigma/2, -r1[0]+d, -r1[1],  r1[2], -r2[0]+d, -r2[1],  r2[2], -r3[0]+d, -r3[1],  r3[2], -nc[0], -nc[1], nc[2], theta[t], -phi[t][0], dsigma_t)))
                    #     Ts.append(np.array((-vc[0]+d, -vc[1],  vc[2], dsigma/2, -r3[0]+d, -r3[1],  r3[2], -r4[0]+d, -r4[1],  r4[2], -r1[0]+d, -r1[1],  r1[2], -nc[0], -nc[1], nc[2], theta[t], -phi[t][0], dsigma_t)))

                    #     Ts.append(np.array((-vc[0]+d,  vc[1], -vc[2], dsigma/2, -r1[0]+d,  r1[1], -r1[2], -r2[0]+d,  r2[1], -r2[2], -r3[0]+d,  r3[1], -r3[2], -nc[0],  nc[1], -nc[2], np.pi-theta[t], phi[t][0], dsigma_t)))
                    #     Ts.append(np.array((-vc[0]+d,  vc[1], -vc[2], dsigma/2, -r3[0]+d,  r3[1], -r3[2], -r4[0]+d,  r4[1], -r4[2], -r1[0]+d,  r1[1], -r1[2], -nc[0],  nc[1], -nc[2], np.pi-theta[t], phi[t][0], dsigma_t)))

                    #     Ts.append(np.array((-vc[0]+d, -vc[1], -vc[2], dsigma/2, -r1[0]+d, -r1[1], -r1[2], -r2[0]+d, -r2[1], -r2[2], -r3[0]+d, -r3[1], -r3[2], -nc[0], -nc[1], -nc[2], np.pi-theta[t], -phi[t][0], dsigma_t)))
                    #     Ts.append(np.array((-vc[0]+d, -vc[1], -vc[2], dsigma/2, -r3[0]+d, -r3[1], -r3[2], -r4[0]+d, -r4[1], -r4[2], -r1[0]+d, -r1[1], -r1[2], -nc[0], -nc[1], -nc[2], np.pi-theta[t], -phi[t][0], dsigma_t)))


    # Assemble a mesh table:
    table = np.array(Ts)
    return table



from phoebe import u

def unit_choices(unit):
    """
    Return a (non-complete) list of available units of the same physical type.

    Arguments
    -----------
    * `unit` (astropy.Unit object or str): unit to be converted

    Returns
    ----------
    * (list of strings) list of available units with the same physical type
    """
    if isinstance(unit, str):
        unit = u.Unit(unit)

    physical_type = unit.physical_type

    if physical_type == 'dimensionless':
        if unit.to_string() in ['', 'dimensionless']:
            return ['dimensionless']
        else:
            physical_types = [un.physical_type for un in unit.bases]
            if all([pt=='time' for pt in physical_types]):
                # then we're time/time, eg. s / yr
                # NOTE: doing same units on numerator and denominator will force this to be dimensionless
                # and then we won't know that its time at all.
                return ['d / yr', 'h / d', 'h / yr', 'min / d', 'min / h', 'min / yr', 's / min', 's / h', 's / d', 's / yr']
            elif physical_types == ['power', 'length']:
                # then assume flux units
                return ['W / m2', 'W / km2', 'W / pc2', 'W / AU2', 'W / solRad2', 'solLum / m2', 'solLum / km2', 'solLum / pc2', 'solLum / AU2', 'solLum / solRad2']
            else:
                return [unit.to_string()]
    elif physical_type == 'time':
        return ['yr', 'd', 'h', 'min', 's']
    elif physical_type == 'mass':
        return ['solMass', 'jupiterMass', 'earthMass', 'kg', 'g']
    elif physical_type == 'length':
        return ['solRad', 'jupiterRad', 'earthRad', 'AU', 'pc', 'km', 'm', 'cm', 'mm', 'um', 'nm', 'angstrom']
    elif physical_type == 'area':
        return ['solRad2', 'jupiterRad2', 'earthRad2', 'AU2', 'pc2', 'km2', 'm2', 'cm2']
    elif physical_type == 'volume':
        return ['solRad3', 'jupiterRad3', 'earthRad3', 'AU3', 'pc3', 'km3', 'm3', 'cm3']
    elif physical_type == 'speed':
        return ['solRad / d', 'solRad / s', 'km / h', 'km / s', 'm / h', 'm / s']
    elif physical_type == 'angle':
        if unit.to_string() == 'cycle':
            return ['cycle']
        return ['rad', 'deg']
    elif physical_type == 'angular speed':
        return ['rad / yr', 'rad / d', 'rad / h', 'rad / s', 'deg / yr', 'deg / d', 'deg / h', 'deg / s']
    elif physical_type == 'temperature':
        # astropy cannot convert to u.deg_C or u.imperial.deg_F
        # could support with equivalencies: https://docs.astropy.org/en/stable/units/equivalencies.html#temperature-equivalency
        # but then we'd need to manually pass that within PHOEBE whenever
        # the physical type is temperature
        return ['K']
    elif physical_type == 'power':
        return ['W', 'solLum']
    else:
        return [unit.to_string()]



import logging
import numpy as np
from phoebe import dynamics

from scipy.optimize import newton

logger = logging.getLogger("ETVS")

def barycentric():
    """
    """
    raise NotImplementedError


def crossing(b, component, time, dynamics_method='keplerian', ltte=True, tol=1e-4, maxiter=1000):
    """
    tol in days
    """


    def projected_separation_sq(time, b, dynamics_method, cind1, cind2, ltte=True):
        """
        """
        #print "*** projected_separation_sq", time, dynamics_method, cind1, cind2, ltte


        times = np.array([time])

        if dynamics_method in ['nbody', 'rebound']:
            # TODO: make sure that this takes systemic velocity and corrects positions and velocities (including ltte effects if enabled)
            ts, xs, ys, zs, vxs, vys, vzs = dynamics.nbody.dynamics_from_bundle(b, times, compute=None, ltte=ltte)

        elif dynamics_method=='bs':
            ts, xs, ys, zs, vxs, vys, vzs = dynamics.nbody.dynamics_from_bundle_bs(b, times, compute, ltte=ltte)

        elif dynamics_method=='keplerian':
            # TODO: make sure that this takes systemic velocity and corrects positions and velocities (including ltte effects if enabled)
            ts, xs, ys, zs, vxs, vys, vzs = dynamics.keplerian.dynamics_from_bundle(b, times, compute=None, ltte=ltte, return_euler=False)

        else:
            raise NotImplementedError


        return (xs[cind2][0]-xs[cind1][0])**2 + (ys[cind2][0]-ys[cind1][0])**2


    # TODO: optimize this by allowing to pass cind1 and cind2 directly (and fallback to this if they aren't)
    starrefs = b.hierarchy.get_stars()
    cind1 = starrefs.index(component)
    cind2 = starrefs.index(b.hierarchy.get_sibling_of(component))

    # TODO: provide options for tol and maxiter (in the frontend computeoptionsp)?
    return newton(projected_separation_sq, x0=time, args=(b, dynamics_method, cind1, cind2, ltte), tol=tol, maxiter=maxiter)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from astropy import __version__ as astropyversion

# import these so they'll be available as unitsiau2015.u and unitsiau2015.c
import astropy.constants as c
import astropy.units as u

from astropy.units import def_unit
from astropy.constants import Constant

def _register_unit(unit):
    """
    register this unit in the phoebe.u namespace
    """
    for name in unit._names:
        setattr(u, name, unit)
        for reg in u.core._unit_registries:
            reg._registry[name] = unit


# TODO: pass namespace for units package so that prefixes are loaded (otherwise calling ksolMass will call from the old constants)
ns = None

from distutils.version import LooseVersion
if LooseVersion(astropyversion) < LooseVersion('2.0'):


    """
    see https://www.iau.org/static/resolutions/IAU2015_English.pdf

    Here we'll override astropy's constants to adhere to the IAU Resolution for nominal units
    """

    # TODO: find correct error estimate for this value of G
    G = Constant('G', "Gravitational constant", 6.67408e-11, 'm3 / (kg s2)', 0.00080e-11, 'NSFA 2011', system='si')
    c.G = G
    c.si.G = G

    GM_sun = Constant('GM_sun', "Solar G*M", 1.3271244e20,  'm3 / (s2)', None, "IAU 2015 Resolution B3", system='si')
    c.GM_sun = GM_sun
    c.si.GM_sun = GM_sun

    GM_earth = Constant('GM_earth', "Earth G*M", 3.986004e14, 'm3 / (s2)', None, "IAU 2015 Resolution B3", system='si')
    c.GM_earth = GM_earth
    c.si.GM_earth = GM_earth

    GM_jup = Constant('GM_jup', "Juptiter G*M", 1.2668653e17, 'm3 / (s2)', None, "IAU 2015 Resolution B3", system='si')
    c.GM_jup = GM_jup
    c.si.GM_jup = GM_jup

    M_sun = Constant('M_sun', "Solar mass", c.GM_sun.value/c.G.value, 'kg', None, "IAU 2015 Resolution B3 (derived from GM_sun and G)", system='si')
    c.M_sun = M_sun
    c.si.M_sun = M_sun

    R_sun = Constant('R_sun', "Solar radius", 6.957e8, 'm', None, "IAU 2015 Resolution B3", system='si')
    c.R_sun = R_sun
    c.si.R_sun = R_sun

    L_sun = Constant('L_sun', "Solar luminosity", 3.828e26, 'W', None, "IAU 2015 Resolution B3", system='si')
    c.L_sun = L_sun
    c.si.L_sun = L_sun

    """
    Now we need to redefine the units to use these constants
    """


    solMass = def_unit(['solMass', 'M_sun', 'Msun'], c.si.M_sun, namespace=ns,
             prefixes=True, doc="Solar mass (nominal)",
             format={'latex': r'{\mathcal{M}^{\rm N}_\odot}'})

    _register_unit(solMass)


    solRad = def_unit(['solRad', 'R_sun', 'Rsun'], c.si.R_sun, namespace=ns,
             doc="Solar radius (nominal)", prefixes=True,
             format={'latex': r'{\mathcal{R}^{\rm N}_\odot}'})

    _register_unit(solRad)

    solLum = def_unit(['solLum', 'L_sun', 'Lsun'], c.si.L_sun, namespace=ns,
             prefixes=True, doc="Solar luminance (nominal)",
             format={'latex': r'{\mathcal{L}^{\rm N}_\odot}'})

    _register_unit(solLum)

    """
    Let's cleanup the imports so these entries don't
    show at the top-level
    """

    del G
    del GM_sun
    del M_sun
    del R_sun
    del L_sun
    del solMass
    del solRad
    del solLum


else:
    # current defaults in astropy (as of 2.0) are codata2014 and iau2015
    pass


"""
T_sun is not provided by astropy 2.0, so we'll set the constant and the unit
for any version of astropy
"""

T_sun = Constant('T_sun', "Solar effective temperature", 5772, 'K', None, "IAU 2015 Resolution B3", system='si')
c.T_sun = T_sun
c.si.T_sun = T_sun

solTeff = def_unit(['solTeff', 'T_sun', 'Tsun'], c.si.T_sun, namespace=ns,
        prefixes=True, doc="Solar effective temperature (nominal)",
        format={'latex': r'{\mathcal{T}^{\rm N}_\odot}'})

_register_unit(solTeff)

"""
astropy also doesn't have any ability to convert to solar units, so we'll provide
a function for that.

TODO: eventually consider adopting this as a "base" unit
"""
_physical_types_to_solar = {'length': 'solRad',
                            'mass': 'solMass',
                            'temperature': 'solTeff',
                            'power': 'solLum',
                            'time': 'd',
                            'speed': 'solRad/d',
                            'angle': 'rad',
                            'angular speed': 'rad/d',
                            'dimensionless': ''}

_physical_types_to_si = {'length': 'm',
                            'mass': 'kg',
                            'temperature': 'K',
                            'power': 'W',
                            'time': 's',
                            'speed': 'm/s',
                            'angle': 'rad',
                            'angular speed': 'rad/s',
                            'dimensionless': ''}

def _get_physical_type(object):

    if hasattr(object, 'physical_type'):
        unit = object
    elif isinstance(object, u.Quantity):
        unit = object.unit
    else:
        raise NotImplementedError("object {} with type={} not supported for _get_physical_type".format(object, type(object)))

    return unit.physical_type

def can_convert_to_solar(object):
    return _get_physical_type(object) in _physical_types_to_solar.keys()

def to_solar(object):
    """
    Arguments
    ----------
    * `object` (quantity or unit)

    Returns
    ----------
    * quantity or unit in applicable solar units

    Raises
    --------
    * NotImplementedError: if cannot convert to solar
    """
    physical_type = _get_physical_type(object)

    if physical_type not in _physical_types_to_solar.keys():
        raise NotImplementedError("cannot convert object with physical_type={} to solar units".format(physical_type))

    return object.to(_physical_types_to_solar.get(physical_type))

u.can_convert_to_solar = can_convert_to_solar
u.to_solar = to_solar
u._get_physical_type = _get_physical_type
u._physical_types_to_si  = _physical_types_to_si
u._physical_types_to_solar = _physical_types_to_solar



"""
And lastly, let's do all remaining cleanup
"""
del LooseVersion

del T_sun
del Constant

del solTeff
del ns
del def_unit





import numpy as np
from numpy import pi,sqrt,cos,sin,tan,arctan
from scipy.optimize import newton

from phoebe import u, c
from phoebe import conf

import logging
logger = logging.getLogger("DYNAMICS.KEPLERIAN")
logger.addHandler(logging.NullHandler())

_skip_filter_checks = {'check_default': False, 'check_visible': False}

def dynamics_from_bundle(b, times, compute=None, return_euler=False, **kwargs):
    """
    Parse parameters in the bundle and call :func:`dynamics`.

    See :func:`dynamics` for more detailed information.

    NOTE: you must either provide compute (the label) OR all relevant options
    as kwargs (ltte)

    Args:
        b: (Bundle) the bundle with a set hierarchy
        times: (list or array) times at which to run the dynamics
        return_euler: (bool, default=False) whether to include euler angles
            in the return

    Returns:
        t, xs, ys, zs, vxs, vys, vzs [, theta, longan, incl].
        t is a numpy array of all times,
        the remaining are a list of numpy arrays (a numpy array per
        star - in order given by b.hierarchy.get_stars()) for the cartesian
        positions and velocities of each star at those same times.
        Euler angles (theta, longan, incl) are only returned if return_euler is
        set to True.

    """

    b.run_delayed_constraints()

    computeps = b.get_compute(compute=compute, force_ps=True, **_skip_filter_checks)
    if len(computeps.computes) == 1:
        ltte = computeps.get_value(qualifier='ltte', ltte=kwargs.get('ltte', None), default=False, **_skip_filter_checks)
    else:
        ltte = False

    # make sure times is an array and not a list
    times = np.array(times)

    vgamma = b.get_value(qualifier='vgamma', context='system', unit=u.solRad/u.d, **_skip_filter_checks)
    t0 = b.get_value(qualifier='t0', context='system', unit=u.d, **_skip_filter_checks)

    hier = b.hierarchy
    starrefs = hier.get_stars()
    orbitrefs = hier.get_orbits()
    s = b.filter(context='component', **_skip_filter_checks)

    periods, eccs, smas, t0_perpasses, per0s, long_ans, incls, dpdts, \
    deccdts, dperdts, components = [],[],[],[],[],[],[],[],[],[],[]


    for component in starrefs:

        # we need to build a list of all orbitlabels underwhich this component
        # belongs.  For a simple binary this is just the parent, but for hierarchical
        # systems we need to get the labels of the outer-orbits as well
        ancestororbits = []
        comp = component
        while hier.get_parent_of(comp) in orbitrefs:
            comp = hier.get_parent_of(comp)
            ancestororbits.append(comp)

        #print "***", component, ancestororbits

        periods.append([s.get_value(qualifier='period_anom', unit=u.d, component=orbit, **_skip_filter_checks) for orbit in ancestororbits])
        eccs.append([s.get_value(qualifier='ecc', component=orbit, **_skip_filter_checks) for orbit in ancestororbits])
        t0_perpasses.append([s.get_value(qualifier='t0_perpass', unit=u.d, component=orbit, **_skip_filter_checks) for orbit in ancestororbits])
        per0s.append([s.get_value(qualifier='per0', unit=u.rad, component=orbit, **_skip_filter_checks) for orbit in ancestororbits])
        long_ans.append([s.get_value(qualifier='long_an', unit=u.rad, component=orbit, **_skip_filter_checks) for orbit in ancestororbits])
        incls.append([s.get_value(qualifier='incl', unit=u.rad, component=orbit, **_skip_filter_checks) for orbit in ancestororbits])
        # TODO: do we need to convert from anomalistic to sidereal?
        dpdts.append([s.get_value(qualifier='dpdt', unit=u.d/u.d, component=orbit, **_skip_filter_checks) for orbit in ancestororbits])
        if conf.devel:
            try:
                deccdts.append([s.get_value('deccdt', u.dimensionless_unscaled/u.d, component=orbit, **_skip_filter_checks) for orbit in ancestororbits])
            except ValueError:
                deccdts.append([0.0 for orbit in ancestororbits])
        else:
            deccdts.append([0.0 for orbit in ancestororbits])
        dperdts.append([s.get_value(qualifier='dperdt', unit=u.rad/u.d, component=orbit, **_skip_filter_checks) for orbit in ancestororbits])

        # sma needs to be the COMPONENT sma.  This is stored in the bundle for stars, but is NOT
        # for orbits in orbits, so we'll need to recompute those from the mass-ratio and sma of
        # the parent orbit.

        smas_this = []
        for comp in [component]+ancestororbits[:-1]:
            if comp in starrefs:
                smas_this.append(s.get_value(qualifier='sma', unit=u.solRad, component=comp, **_skip_filter_checks))
            else:
                q = s.get_value(qualifier='q', component=hier.get_parent_of(comp), **_skip_filter_checks)
                comp_comp = hier.get_primary_or_secondary(comp)

                # NOTE: similar logic is also in constraints.comp_sma
                # If changing any of the logic here, it should be changed there as well.
                if comp_comp == 'primary':
                    qthing = (1. + 1./q)
                else:
                    qthing = (1. + q)

                smas_this.append(s.get_value(qualifier='sma', unit=u.solRad, component=hier.get_parent_of(comp), **_skip_filter_checks) / qthing)

        smas.append(smas_this)

        # components is whether an entry is the primary or secondary in its parent orbit, so here we want
        # to start with component and end one level short of the top-level orbit
        components.append([hier.get_primary_or_secondary(component=comp) for comp in [component]+ancestororbits[:-1]])


    return  dynamics(times, periods, eccs, smas, t0_perpasses, per0s, \
                    long_ans, incls, dpdts, deccdts, dperdts, \
                    components, t0, vgamma, \
                    mass_conservation=True, ltte=ltte, return_euler=return_euler)




def dynamics(times, periods, eccs, smas, t0_perpasses, per0s, long_ans, incls,
            dpdts, deccdts, dperdts, components, t0=0.0, vgamma=0.0,
            mass_conservation=True, ltte=False, return_euler=False):
    """
    Compute the positions and velocites of each star in their nested
    Keplerian orbits at a given list of times.

    See :func:`dynamics_from_bundle` for a wrapper around this function
    which automatically handles passing everything in the correct order
    and in the correct units.

    Args:
        times: (iterable) times at which to compute positions and
            velocities for each star
        periods: (iterable) anomalistic period of the parent orbit for each star
            [days]
        eccs: (iterable) eccentricity of the parent orbit for each star
        smas: (iterable) semi-major axis of the parent orbit for each
            star [solRad]
        t0_perpasses: (iterable) t0_perpass of the parent orbit for each
            star [days]
        per0s: (iterable) longitudes of periastron of the parent orbit
            for each star [rad]
        long_ans: (iterable) longitudes of the ascending node of the
            parent orbit for each star [rad]
        incls: (iterable) inclination of the parent orbit for each
            star [rad]
        dpdts: (iterable) change in period with respect to time of the
            parent orbit for each star [days/day]
        deccdts: (iterable) change in eccentricity with respect to time
            of the parent orbit for each star [1/day]
        dperdts: (iterable) change in periastron with respect to time
            of the parent orbit for each star [rad/d]
        components: (iterable) component ('primary' or 'secondary') of
            each star within its parent orbit [string]
        t0: (float, default=0) time at which all initial values (ie period, per0)
            are given [days]
        mass_conservation: (bool, optional) whether to require mass
            conservation if any of the derivatives (dpdt, dperdt, etc)
            are non-zero [default: True]
        return_euler: (bool, default=False) whether to include euler angles
            in the return

    Returns:
        t, xs, ys, zs, vxs, vys, vzs [, theta, longan, incl].
        t is a numpy array of all times,
        the remaining are a list of numpy arrays (a numpy array per
        star - in order given by b.hierarchy.get_stars()) for the cartesian
        positions and velocities of each star at those same times.
        Euler angles (theta, longan, incl) are only returned if return_euler is
        set to True.
    """
    # TODO: NOTE: smas must be per-component, not per-orbit
    # TODO: steal some documentation from 2.0a:keplerorbit.py:get_orbit
    # TODO: deal with component number more smartly

    def binary_dynamics(times, period, ecc, sma, t0_perpass, per0, long_an,
                        incl, dpdt, deccdt, dperdt, component='primary',
                        t0=0.0, vgamma=0.0, mass_conservation=True,
                        com_pos=(0.,0.,0.), com_vel=(0.,0.,0.), com_euler=(0.,0.,0.)):
        """
        """
        # TODO: steal some documentation from 2.0a:keplerorbit.py:get_orbit


        #-- if dpdt is non-zero, the period is actually an array, and the semi-
        #   major axis changes to match Kepler's third law (unless
        #   `mass_conservation` is set to False)
        # p0 = period
        # sma0 = sma

        if dpdt!=0:
            p0 = period
            period = dpdt*(times-t0) + p0
            if mass_conservation:
                # Pieter used to have a if not np.isscale(period) to use period[0] instead of p0,
                # effectively starting mass conservation at the first dataset time instead of t0
                sma = sma/p0**2*period**2

        #-- if dperdt is non-zero, the argument of periastron is actually an
        #   array
        if dperdt!=0.:
            per0 = dperdt*(times-t0) + per0
        #-- if deccdt is non-zero, the eccentricity is actually an array
        if deccdt!=0:
            ecc = deccdt*(times-t0) + ecc


        #-- compute orbit
        n = 2*pi/period
        ma = n*(times-t0_perpass)
        E,theta = _true_anomaly(ma,ecc)
        r = sma*(1-ecc*cos(E))
        PR = r*sin(theta)
        #-- compute rdot and thetadot
        l = r*(1+ecc*cos(theta))#-omega))
        L = 2*pi*sma**2/period*sqrt(1-ecc**2)
        rdot = L/l*ecc*sin(theta)#-omega)
        thetadot = L/r**2
        #-- the secondary is half an orbit further than the primary
        if 'sec' in component.lower():
            theta += pi
        #-- take care of Euler angles
        theta_ = theta+per0
        #-- convert to the right coordinate frame
        #-- create some shortcuts
        sin_theta_ = sin(theta_)
        cos_theta_ = cos(theta_)
        sin_longan = sin(long_an)
        cos_longan = cos(long_an)
        #-- spherical coordinates to cartesian coordinates. Note that we actually
        #   set incl=-incl (but of course it doesn't show up in the cosine). We
        #   do this to match the convention that superior conjunction (primary
        #   eclipsed) happens at periastron passage when per0=90 deg.
        x = r*(cos_longan*cos_theta_ - sin_longan*sin_theta_*cos(incl))
        y = r*(sin_longan*cos_theta_ + cos_longan*sin_theta_*cos(incl))
        z = r*(sin_theta_*sin(-incl))
        #-- spherical vectors to cartesian vectors, and then rotated for
        #   the Euler angles Omega and i.
        vx_ = cos_theta_*rdot - sin_theta_*r*thetadot
        vy_ = sin_theta_*rdot + cos_theta_*r*thetadot
        vx = cos_longan*vx_ - sin_longan*vy_*cos(incl)
        vy = sin_longan*vx_ + cos_longan*vy_*cos(incl)
        vz = sin(-incl)*vy_
        #-- that's it!

        # correct by vgamma (only z-direction)
        # NOTE: vgamma is in the direction of positive RV or negative vz
        vz -= vgamma
        z -= vgamma * (times-t0)

        return (x+com_pos[0],y+com_pos[1],z+com_pos[2]),\
                (vx+com_vel[0],vy+com_vel[1],vz+com_vel[2]),\
                (theta_,long_an,incl)
                # (theta_+com_euler[0],long_an+com_euler[1],incl+com_euler[2])

    def binary_dynamics_nested(times, periods, eccs, smas, \
                            t0_perpasses, per0s, long_ans, incls, dpdts, deccdts, \
                            dperdts, components, t0, vgamma, \
                            mass_conservation):

        """
        compute the (possibly nested) positions of a single component (ltte should be
        handle externally)
        """


        if not hasattr(periods, '__iter__'):
            # then we don't have to worry about nesting, and each item should
            # be a single value ready to pass to binary_dynamics
            pos, vel, euler = binary_dynamics(times, periods, eccs, smas, t0_perpasses, \
                                per0s, long_ans, incls, dpdts, deccdts, dperdts, components, \
                                t0, vgamma, mass_conservation)

        else:
            # Handle nesting - if this star is not in the top-level orbit, then
            # all values should actually be lists.  We want to sort by period to handle
            # the outer orbits first and then apply those offsets to the inner-orbit(s)

            # let's convert to arrays so we can use argsort easily
            periods = np.array(periods)
            eccs = np.array(eccs)
            smas = np.array(smas)
            t0_perpasses = np.array(t0_perpasses)
            per0s = np.array(per0s)
            long_ans = np.array(long_ans)
            incls = np.array(incls)
            dpdts = np.array(dpdts)
            deccdts = np.array(deccdts)
            dperdts = np.array(dperdts)
            components = np.array(components)

            si = periods.argsort()[::-1]

            #print "***", periods, si

            pos = (0.0, 0.0, 0.0)
            vel = (0.0, 0.0, 0.0)
            euler = (0.0, 0.0, 0.0)

            for period, ecc, sma, t0_perpass, per0, long_an, incl, dpdt, \
                deccdt, dperdt, component in zip(periods[si], eccs[si], \
                smas[si], t0_perpasses[si], per0s[si], long_ans[si], \
                incls[si], dpdts[si], deccdts[si], dperdts[si], components[si]):

                pos, vel, euler = binary_dynamics(times, period, ecc, sma, t0_perpass, \
                                    per0, long_an, incl, dpdt, deccdt, dperdt, component, \
                                    t0, vgamma, mass_conservation,
                                    com_pos=pos, com_vel=vel, com_euler=euler)



        return pos, vel, euler




    xs, ys, zs = [], [], []
    vxs, vys, vzs = [], [], []

    if return_euler:
        ethetas, elongans, eincls = [], [], []

    for period, ecc, sma, t0_perpass, per0, long_an, incl, dpdt, deccdt, \
            dperdt, component in zip(periods, eccs, smas, t0_perpasses, per0s, long_ans, \
            incls, dpdts, deccdts, dperdts, components):

        # We now have the orbital parameters for a single star/component.

        if ltte:
            #scale_factor = 1.0/c.c.value * c.R_sun.value/(24*3600.)
            scale_factor = (c.R_sun/c.c).to(u.d).value

            def propertime_barytime_residual(t, time):
                pos, vel, euler = binary_dynamics_nested(t, period, ecc, sma, \
                                t0_perpass, per0, long_an, incl, dpdt, deccdt, \
                                dperdt, components=component, t0=t0, vgamma=vgamma, \
                                mass_conservation=mass_conservation)
                z = pos[2]
                return t - z*scale_factor - time

            # Finding that right time is easy with a Newton optimizer:
            propertimes = [newton(propertime_barytime_residual, time, args=(time,)) for \
               time in times]
            propertimes = np.array(propertimes).ravel()

            pos, vel, euler = binary_dynamics_nested(propertimes, period, ecc, sma, \
                            t0_perpass, per0, long_an, incl, dpdt, deccdt, \
                            dperdt, components=component, t0=t0, vgamma=vgamma, \
                            mass_conservation=mass_conservation)


        else:
            pos, vel, euler = binary_dynamics_nested(times, period, ecc, sma, \
                    t0_perpass, per0, long_an, incl, dpdt, deccdt, \
                    dperdt, components=component, t0=t0, vgamma=vgamma, \
                    mass_conservation=mass_conservation)


        xs.append(pos[0])
        ys.append(pos[1])
        zs.append(pos[2])
        vxs.append(vel[0])
        vys.append(vel[1])
        vzs.append(vel[2])
        if return_euler:
            ethetas.append(euler[0])
            elongans.append([euler[1]]*len(euler[0]))
            eincls.append([euler[2]]*len(euler[0]))


    # if return_euler:
    #     return times, \
    #         xs*u.solRad, ys*u.solRad, zs*u.solRad, \
    #         vxs*u.solRad/u.d, vys*u.solRad/u.d, vzs*u.solRad/u.d, \
    #         ethetas*u.rad, elongans*u.rad, eincls*u.rad
    # else:
    #     return times, \
    #         xs*u.solRad, ys*u.solRad, zs*u.solRad, \
    #         vxs*u.solRad/u.d, vys*u.solRad/u.d, vzs*u.solRad/u.d    if return_euler:

    # d, solRad, solRad/d, rad
    if return_euler:
        return times, \
            xs, ys, zs, \
            vxs, vys, vzs, \
            ethetas, elongans, eincls
    else:
        return times, \
            xs, ys, zs, \
            vxs, vys, vzs




def _true_anomaly(M,ecc,itermax=8):
    r"""
    Calculation of true and eccentric anomaly in Kepler orbits.

    ``M`` is the phase of the star, ``ecc`` is the eccentricity

    See p.39 of Hilditch, 'An Introduction To Close Binary Stars':

    Kepler's equation:

    .. math::

        E - e\sin E = \frac{2\pi}{P}(t-T)

    with :math:`E` the eccentric anomaly. The right hand size denotes the
    observed phase :math:`M`. This function returns the true anomaly, which is
    the position angle of the star in the orbit (:math:`\theta` in Hilditch'
    book). The relationship between the eccentric and true anomaly is as
    follows:

    .. math::

        \tan(\theta/2) = \sqrt{\frac{1+e}{1-e}} \tan(E/2)

    @parameter M: phase
    @type M: float
    @parameter ecc: eccentricity
    @type ecc: float
    @keyword itermax: maximum number of iterations
    @type itermax: integer
    @return: eccentric anomaly (E), true anomaly (theta)
    @rtype: float,float
    """
    # Initial value
    Fn = M + ecc*sin(M) + ecc**2/2.*sin(2*M)

    # Iterative solving of the transcendent Kepler's equation
    for i in range(itermax):
        F = Fn
        Mn = F-ecc*sin(F)
        Fn = F+(M-Mn)/(1.-ecc*cos(F))
        keep = F!=0 # take care of zerodivision
        if hasattr(F,'__iter__'):
            if np.all(abs((Fn-F)[keep]/F[keep])<0.00001):
                break
        elif (abs((Fn-F)/F)<0.00001):
            break

    # relationship between true anomaly (theta) and eccentric anomaly (Fn)
    true_an = 2.*arctan(sqrt((1.+ecc)/(1.-ecc))*tan(Fn/2.))

    return Fn,true_an



import logging

import sys

import numpy as np

_skip_filter_checks = {'check_default': False, 'check_visible': False}

import logging
logger = logging.getLogger("UTILS")
logger.addHandler(logging.NullHandler())

def _bytes(s):
    return bytes(s, 'utf-8')

def get_basic_logger(clevel='WARNING',flevel='DEBUG',
                     style="default",filename=None,filemode='w'):
    """
    Return a basic logger via a log file and/or terminal.

    Example 1: log only to the console, accepting levels "INFO" and above

    >>> logger = utils.get_basic_logger()

    Example 2: log only to the console, accepting levels "DEBUG" and above

    >>> logger = utils.get_basic_logger(clevel='DEBUG')

    Example 3: log only to a file, accepting levels "DEBUG" and above

    >>> logger = utils.get_basic_logger(clevel=None,filename='mylog.log')

    Example 4: log only to a file, accepting levels "INFO" and above

    >>> logger = utils.get_basic_logger(clevel=None,flevel='INFO',filename='mylog.log')

    Example 5: log to the terminal (INFO and above) and file (DEBUG and above)

    >>> logger = utils.get_basic_logger(filename='mylog.log')

    The different logging styles are:

    C{style='default'}::

        Wed, 13 Feb 2013 08:47 root INFO     Some information

    C{style='grandpa'}::

        # INFO    Some information

    C{style='minimal'}::

        Some information

    @param style: logger style
    @type style: str, one of 'default','grandpa','minimal'
    """
    name = ""
    #-- define formats
    if style=='default':
        format  = '%(asctime)s %(name)-12s %(levelname)-7s %(message)s'
        datefmt = '%a, %d %b %Y %H:%M'
    elif style=='grandpa':
        format  = '# %(levelname)-7s %(message)s'
        datefmt = '%a, %d %b %Y %H:%M'
    elif style=='minimal':
        format = ''
        datefmt = '%a, %d %b %Y %H:%M'

    if style=='trace':
        formatter = MyFormatter()
    else:
        formatter = logging.Formatter(fmt=format,datefmt=datefmt)


    if clevel: clevel = logging.__dict__[clevel.upper()]
    if flevel: flevel = logging.__dict__[flevel.upper()]

    #-- set up basic configuration.
    #   The basicConfig sets up one default logger. If you give a filename, it's
    #   a FileHandler, otherwise a StreamHandler.
    #-- If we want console and filename, first set up a basic FileHandler, then
    #   add terminal StreamHandler
    if filename is not None:
        if flevel is None:
            level = clevel
        else:
            level = flevel
        logging.basicConfig(level=level,
                            format=format,datefmt=datefmt,
                            filename=filename,filemode=filemode)
        fh = logging.FileHandler(filename)
        fh.setLevel(flevel)
        fh.setFormatter(formatter)
        logging.getLogger(name).addHandler(fh)
    if filename is not None and clevel:
        # define a Handler which writes INFO messages or higher to the sys.stderr
        ch = logging.StreamHandler()
        ch.setLevel(clevel)
        # tell the handler to use this format
        ch.setFormatter(formatter)
        logging.getLogger(name).addHandler(ch)
    #-- If we only want a console:
    else:
        logging.basicConfig(level=clevel,format=format,datefmt=datefmt,
                            filename=filename,filemode=filemode)
    #-- fix filename logging
    if filename is not None:
        logging.getLogger(name).handlers[0].level = flevel
    return logging.getLogger(name)

def pop_streamhandlers(logger):
    #-- remove streamhandlers -- they don't have a baseFilename
    to_remove = [handler for handler in logger.handlers if not hasattr(handler,'baseFilename')]
    logger.handlers = [handler for handler in logger.handlers if hasattr(handler,'baseFilename')]
    return to_remove

def pop_filehandlers(logger):
    #-- remove filehandlers -- they have a baseFilename
    to_remove = [handler for handler in logger.handlers if hasattr(handler,'baseFilename')]
    logger.handlers = [handler for handler in logger.handlers if not hasattr(handler,'baseFilename')]
    return to_remove

def add_filehandler(logger,style="default",flevel='DEBUG',
                   filename=None,filemode='w'):
    name = ""
    #-- define formats
    if style=='default':
        format  = '%(asctime)s %(name)-12s %(levelname)-7s %(message)s'
        datefmt = '%a, %d %b %Y %H:%M'
    elif style=='grandpa':
        format  = '# %(levelname)-7s %(message)s'
        datefmt = '%a, %d %b %Y %H:%M'
    elif style=='minimal':
        format = ''
        datefmt = '%a, %d %b %Y %H:%M'
    flevel = logging.__dict__[flevel.upper()]
    logging.basicConfig(level=flevel,
                        format=format,datefmt=datefmt,
                        filename=filename,filemode=filemode)
    fh = logging.FileHandler(filename)
    fh.setLevel(flevel)
    formatter = logging.Formatter(fmt=format,datefmt=datefmt)
    fh.setFormatter(formatter)
    logging.getLogger(name).addHandler(fh)
    logging.getLogger(name).handlers[-1].level = flevel
    return logging.getLogger(name)

def parse_json(pairs):
    """
    modified from:
    https://stackoverflow.com/questions/956867/how-to-get-string-objects-instead-of-unicode-from-json#34796078

    pass this to the object_pairs_hook kwarg of json.load/loads
    """
    def _string(item):
        if isinstance(item, bytes):
            # return item.decode('utf-8')
            return _bytes(item)
        else:
            return item

    new_pairs = []
    for key, value in pairs:
        key = _string(key)

        if isinstance(value, dict):
            value = parse_json(value.items())
        elif isinstance(value, list):
            value = [_string(v) for v in value]
        else:
            value = _string(value)

        new_pairs.append((key, value))
    return dict(new_pairs)

def phase_mask_inds(phases, mask_phases):
    def _individual_mask(phases, mask):
        # move mask onto range (-0.5, 0.5)
        def _map(m):
            if m < -0.5:
                return _map(m+1)
            if m > 0.5:
                return _map(m-1)
            return m

        mask = [_map(m) for m in mask]

        does_wrap = mask[0] > mask[1]

        if does_wrap:
            return np.logical_or(phases > mask[0], phases < mask[1])
        else:
            return np.logical_and(phases >= mask[0], phases <= mask[1])

    if mask_phases is None:
        return np.isfinite(phases)

    masks = [_individual_mask(phases, m) for m in mask_phases]
    if len(masks) == 0:
        inds = np.isfinite(phases)
    elif len(masks) == 1:
        inds = masks[0]
    else:
        inds = np.logical_or(*masks)

    return inds


def _get_masked_times(b, dataset, mask_phases, mask_t0, return_times_phases=False):
    # concatenate for the case of datasets (like RVs) with times in multiple components
    times = np.unique(np.concatenate([time_param.get_value() for time_param in b.filter(qualifier='times', dataset=dataset, **_skip_filter_checks).to_list()]))
    phases = b.to_phases(times, t0=mask_t0)
    masked_times = times[phase_mask_inds(phases, mask_phases)]
    if return_times_phases:
        return masked_times, times, phases
    return masked_times

def _get_masked_compute_times(b, dataset, mask_phases, mask_t0, is_time_dependent, times=None, phases=None):
    # for compute_times/phases we can't just mask because we need to make
    # sure we "surround" each of the observation datapoints
    if times is None:
        times = np.unique(np.concatenate([time_param.get_value() for time_param in b.filter(qualifier='times', dataset=dataset, unit='d', **_skip_filter_checks).to_list()]))
    if phases is None:
        phases = b.to_phases(times, t0=mask_t0)

    compute_times = b.get_value(qualifier='compute_times', dataset=dataset, context='dataset', unit='d', **_skip_filter_checks)

    if mask_phases is None:
        return compute_times

    compute_phases = b.to_phases(compute_times, t0=mask_t0)

    indices = []

    def _phase_diff(ph1, ph2):
        # need to account for phase-wrapping when finding the nearest two points
        return min([abs(ph1-ph2), abs(ph1+1-ph2), abs(ph1-1-ph2)])

    if is_time_dependent:
        times_masked = times[phase_mask_inds(phases, mask_phases)]

        for tm in times_masked:
            indices += list(abs(compute_times-tm).argsort()[:2])
    else:
        phases_masked = phases[phase_mask_inds(phases, mask_phases)]

        for phm in phases_masked:
            indices += list(np.array([_phase_diff(cph, phm) for cph in compute_phases]).argsort()[:2])

    return compute_times[list(set(indices))]




from phoebe.parameters import *
from phoebe import u

### NOTE: if creating new parameters, add to the _forbidden_labels list in parameters.py

def system(**kwargs):
    """
    Generally, this will automatically be added to a newly initialized
    <phoebe.frontend.bundle.Bundle>.

    Arguments
    -----------
    * `t0` (float/quantity, optional): time at which all values are defined.
    * `ra` (float/quantity, optional): right ascension.
    * `dec` (float/quantity, optiona): declination.
    * `epoch` (string, optional): epoch of `ra` and `dec`.
    * `distance` (float/quantity, optional): distance to the system.
    * `vgamma` (float/quantity, optional): Constant barycentric systemic velocity
        (in the direction of positive RV or negative vz)
    * `ebv` (float, optional, default=0): extinction E(B-V).
    * `Av` (float, optional, default=0): extinction Av.
    * `Rv` (float, optional, default=3.1): extinction law parameter.

    Returns
    --------
    * (<phoebe.parameters.ParameterSet) ParameterSet of all created Parameters.
    """
    params = []
    constraints = []

    params += [FloatParameter(qualifier='t0', latexfmt=r't_0', value=kwargs.get('t0', 0.0), default_unit=u.d, description='Time at which all values are provided.  For values with time-derivatives, this defines their zero-point.')]

    # TODO: re-enable these once they're incorporated into orbits (dynamics) correctly.
    params += [FloatParameter(qualifier='ra', latexfmt=r'\alpha', value=kwargs.get('ra', 0.0), default_unit=u.deg, advanced=True, description='Right ascension')]
    params += [FloatParameter(qualifier='dec', latexfmt=r'\delta', value=kwargs.get('dec', 0.0), default_unit=u.deg, advanced=True, description='Declination')]
    # params += [StringParameter(qualifier='epoch', value=kwargs.get('epoch', 'J2000'), advanced=True, description='Epoch of coordinates')]
    #params += [FloatParameter(qualifier='pmra', value=kwargs.get('pmra', 0.0), default_unit=u.mas/u.yr, description='Proper motion in right ascension')]
    #params += [FloatParameter(qualifier='pmdec', value=kwargs.get('pmdec', 0.0), default_unit=u.mas/u.yr, description='Proper motion in declination')]

    params += [FloatParameter(qualifier='distance', latexfmt=r'd', value=kwargs.get('distance', 1.0), default_unit=u.m, description='Distance to the system')]
    params += [FloatParameter(qualifier='vgamma', latexfmt=r'v_\gamma', value=kwargs.get('vgamma', 0.0), default_unit=u.km/u.s, description='Constant barycentric systemic velocity (in the direction of positive RV or negative vz)')]

    params += [FloatParameter(qualifier='ebv', latexfmt=r'E(B-V)', value=kwargs.get('ebv', 0.0), default_unit=u.dimensionless_unscaled, limits=(None, None), description='Extinction E(B-V)')]
    params += [FloatParameter(qualifier='Av', latexfmt=r'A_v', value=kwargs.get('Av', 0.0), default_unit=u.dimensionless_unscaled, limits=(None, None), description='Extinction Av')]
    params += [FloatParameter(qualifier='Rv', latexfmt=r'R_v', value=kwargs.get('Rv', 3.1), default_unit=u.dimensionless_unscaled, limits=(None, None), description='Extinction law parameter')]
    constraints +=[(constraint.extinction,)]

    return ParameterSet(params), constraints



import traceback
import numpy as np
import astropy.units as u
import matplotlib.pyplot as plt

from matplotlib import animation

from . import common
from . import callbacks
from . import call as _call
from . import axes as _axes
from . import mpl_animate as _mpl_animate

class Figure(object):
    """
    """
    def __init__(self, *args, **kwargs):
        """
        Create a new <autofig.figure.Figure> object.

        Arguments
        -----------
        * `*args` (<autofig.call.Call> or <autofig.axes.Axes>, optional):
            positional arguments can be <autofig.call.Call> (<autofig.call.Plot>
            or <autofig.call.Mesh>) or <autofig.axes.Axes> objects to attach to
            the <autofig.figure.Figure>
        * `inline` (bool, optional, default=False): whether in inline mode

        Returns
        -----------
        * <autofig.figure.Figure>: the instantiated <autofig.figure.Figure> object.
        """
        self._class = 'Figure' # just to avoid circular import in order to use isinstance

        self._backend_object = None
        self._backend_artists = []
        self._inline = kwargs.pop('inline', False)

        self._axes = []
        self._calls = []

        if len(kwargs.keys()):
            raise ValueError("kwargs: {} not recognized".format(kwargs.keys()))

        for ca in args:
            if isinstance(ca, _axes.Axes):
                self.add_axes(ca)
            elif isinstance(ca, _call.Call):
                self.add_call(ca)
            else:
                raise TypeError("all arguments must be of type Call or Axes, found {}".format(type(ca)))

    def __repr__(self):
        naxes = len(self.axes)
        ncalls = len(self.calls)
        return "<autofig.figure.Figure | {} axes | {} call(s)>".format(naxes, ncalls)

    @classmethod
    def from_dict(cls, dict):
        args = []
        for axd in dict.pop('axes', []):
            args.append(_axes.Axes.from_dict(axd))
        for calld in dict.pop('calls', []):
            args.append(getattr(_call, calld.pop('classname', 'Plot')).from_dict(calld))

        if len(dict.items()):
            raise ValueError("could not recognize remaining content {}".format(dict))

        return cls(*args)

    def to_dict(self, renders=[]):
        """
        Export the current <autofig.figure.Figure> to a json-safe dictionary.

        See also:
        * <autofig.figure.Figure.save>

        Arguments
        -----------
        * `filename` (string): path to save the figure instance.
        * `renders` (list of dictionaries, default=[]): commands to execute
            for rendering when opened by the command-line tool or by passing
            `do_renders` to <autofig.figure.Figure.open>.  The format must
            be a list of dictionaries, where each dictionary must at least have
            'render': 'draw' or 'render': 'animate'.  Any additional key-value
            pairs will be passed as keyword arguments to the respective
            rendering method.


        Returns
        -----------
        * (str) the path of the saved figure instance.
        """
        renders_json_safe = []
        if len(renders):
            for render in renders:
                if render.get('render', None) not in ['draw', 'animate']:
                    raise ValueError("invalid format for render: {}.  Must include render='draw' or render='animate'".format(render))

                renders_json_safe.append({k: common._json_safe(v) for k,v in render.items()})

        return {'axes': [ax.to_dict() for ax in self.axes], 'calls': [c.to_dict() for c in self.calls], 'renders': renders_json_safe}

    @classmethod
    def open(cls, filename, do_renders=False, allow_renders_save=False):
        """
        Open a <autofig.figure.Figure> from a saved file.

        See also:
        * <autofig.figure.Figure.save>

        Arguments
        -----------
        * `filename` (string): path to the saved figure instance
        * `do_renders` (bool, default=False): whether to execute any render
            (ie. draw/animate) statements included in the file.
        * `allow_renders_save` (bool, default=False): whether to allow render
            statements to save images/animations to disk.  Be careful if setting
            this to True from an untrusted source.

        Returns
        ---------
        * the loaded <autofig.figure.Figure> instance.

        Raises
        ----------
        * ValueError: if `do_render` is True but the render statements are invalid.
        """
        dict = common.load(filename)
        renders = dict.pop('renders', [])
        fig = cls.from_dict(dict)

        if do_renders:
            for render in renders:
                render_cmd = render.pop('render', None)
                if render_cmd not in  ['draw', 'animate']:
                    raise ValueError("invalid format for renders, only accepts draw or animate.  Try passing do_renders=False to skip.")

                if not allow_renders_save:
                    save_dump = render.pop('save', None)

                if 'show' not in render.keys() and 'save' not in render.keys():
                    render['show'] = True

                print("calling {} with kwargs {}".format(render_cmd, render))
                getattr(fig, render_cmd)(**render)

        return fig

    def save(self, filename, renders=[]):
        """
        Save the current <autofig.figure.Figure>.  Note: this saves the autofig
        figure object itself, not the image.  To save the image, call
        <autofig.figure.Figure.draw> and pass `save`.

        See also:
        * <autofig.figure.Figure.open>
        * <autofig.figure.Figure.to_dict>

        Arguments
        -----------
        * `filename` (string): path to save the figure instance.
        * `renders` (list of dictionaries, default=[]): commands to execute
            for rendering when opened by the command-line tool or by passing
            `do_renders` to <autofig.figure.Figure.open>.  The format must
            be a list of dictionaries, where each dictionary must at least have
            'render': 'draw' or 'render': 'animate'.  Any additional key-value
            pairs will be passed as keyword arguments to the respective
            rendering method.


        Returns
        -----------
        * (str) the path of the saved figure instance.
        """
        common.save(self.to_dict(renders=renders), filename)


    @property
    def axes(self):
        """
        Access the children <autofig.axes.Axes> of the <autofig.figure.Figure>

        Returns
        ---------
        * <autofig.axes.AxesGroup> of <autofig.axes.Axes> ordered according to
            <autofig.axes.Axes.axorder>
        """
        axorders = [ax.axorder for ax in self._axes]
        axes = [self._axes[i] for i in np.argsort(axorders)]
        return _axes.AxesGroup(axes)

    def add_axes(self, *axes):
        """
        Add one or multiple <autofig.axes.Axes> to the <autofig.figure.Figure>

        Arguments
        -----------
        * `*args` (<autofig.axes.Axes>, optional): each positional argument must
            be an <autofig.axes.Axes> object.

        Returns
        -----------
        * None
        """
        if len(axes) == 0:
            axes = [_axes.Axes()]

        if len(axes) > 1:
            for a in axes:
                self.add_axes(a)
            return

        elif len(axes) == 1:
            axes = axes[0]
            if not isinstance(axes, _axes.Axes):
                raise TypeError("axes must be of type Axes")

            axes._figure = self
            self._axes.append(axes)
            for call in axes.calls:
                self._calls.append(call)

    @property
    def calls(self):
        """
        Access all children <autofig.call.Call>s of the <autofig.figure.Figure>.

        See also:

        * <autofig.figure.Figure.plots>
        * <autofig.figure.Figure.meshes>

        Returns
        -----------
        * <autofig.call.CallGroup> of <autofig.call.Call> objects
        """
        return _call.make_callgroup(self._calls)

    def add_call(self, *calls):
        """
        Add a <autofig.call.Call> to the <autofig.figure.Figure>

        Arguments
        ----------
        * `*calls`: positional arguments must each be of type <autofig.call.Call>

        Returns
        ---------
        * None
        """
        if len(calls) > 1:
            for c in calls:
                self.add_call(c)
            return

        elif len(calls) == 1:
            call = calls[0]
            if not isinstance(call, _call.Call):
                raise TypeError("call must be of type Call")

            # try to add to existing axes in reverse order before making a new one
            for ax in reversed(self.axes):
                consistent, reason = ax.consistent_with_call(call)
                if consistent:
                    break
            else:
                # then no axes were consistent so we must add a new one
                ax = _axes.Axes()
                self.add_axes(ax)

            ax.add_call(call)
            self._calls.append(call)

    def _get_backend_object(self, fig=None, naxes=1):
        if fig is None:
            if self._backend_object:
                fig = self._backend_object
            else:
                fig = plt.gcf()
                fig.clf()

                rows, cols = _axes._determine_grid(naxes)
                fig.set_figwidth(8*cols)
                fig.set_figheight(6*rows)

                self._backend_artists = []

        self._backend_object = fig
        return fig

    def _get_backend_artists(self):
        return self._backend_artists

    @property
    def plots(self):
        """
        Access all children <autofig.call.Plot>s of the <autofig.figure.Figure>.

        See also:

        * <autofig.figure.Figure.calls>
        * <autofig.figure.Figure.fill_betweens>
        * <autofig.figure.Figure.meshes>

        Returns
        -------------
        * <autofig.call.PlotGroup> of all <autofig.call.Plot> objects
        """
        calls = [c for c in self._calls if isinstance(c, _call.Plot)]
        return _call.PlotGroup(calls)

    def plot(self, *args, **kwargs):
        """
        Add a new <autofig.call.Plot> to the <autofig.figure.Figure>.

        See also:

        * <autofig.call.Plot.__init__>

        Arguments
        ----------
        * `*args`: all positional arguments are passed on to
            <autofig.call.Plot.__init__> to initialize the new
            <autofig.call.Plot>.
        * `tight_layout` (bool, optional, default=True): passed to
            <autofig.figure.Figure.draw> if `show` or `save`.  Whether to draw
            with the `tight_layout` option.
        * `draw_title` (bool, optional, default=True): passed to
            <autofig.figure.Figure.draw> if `show` or `save`.  Whether to draw
            the title on the matplotlib axes.
        * `subplot_grid` (None or tuple, optional, default=None): passed to
            <autofig.figure.Figure.draw> if `show` or `save`.  Override the
            subplot locations.
        * `show` (bool, optional, default=False): whether to immediately
            draw and show the resulting matplotlib figure.  If True,
            <autofig.figure.Figure.draw> will be called.
        * `save` (False or string, optional, default=False): the filename
            to save the resulting matplotlib figure, or False to not save.
            If not False, <autofig.figure.Figure.draw> will be called.
        * `**kwargs`: additional keyword arguments are passed on to
            <autofig.call.Plot.__init__> to initialize the new
            <autofig.call.Plot>.
        """

        tight_layout = kwargs.pop('tight_layout', True)
        draw_sidebars = kwargs.pop('draw_sidebars', True)
        draw_title = kwargs.pop('draw_title', True)
        subplot_grid = kwargs.pop('subplot_grid', None)

        show = kwargs.pop('show', False)
        save = kwargs.pop('save', False)

        call = _call.Plot(*args, **kwargs)
        self.add_call(call)

        if show or save:
            self.reset_draw()
            return self.draw(tight_layout=tight_layout,
                             draw_sidebars=draw_sidebars,
                             draw_title=draw_title,
                             subplot_grid=subplot_grid,
                             show=show, save=save)

    @property
    def fill_betweens(self):
        """
        Access all children <autofig.call.FillBetween>s of the <autofig.figure.Figure>.

        See also:

        * <autofig.figure.Figure.calls>
        * <autofig.figure.Figure.plots>
        * <autofig.figure.Figure.meshes>

        Returns
        -------------
        * <autofig.call.CallGroup> of all <autofig.call.FillBetween> objects
        """
        calls = [c for c in self._calls if isinstance(c, _call.FillBetween)]
        return _call.CallGroup(calls)

    def fill_between(self, *args, **kwargs):
        """
        Add a new <autofig.call.FillBetween> to the <autofig.figure.Figure>.

        See also:

        * <autofig.call.FillBetween.__init__>

        Arguments
        ----------
        * `*args`: all positional arguments are passed on to
            <autofig.call.FillBetween.__init__> to initialize the new
            <autofig.call.FillBetween>.
        * `tight_layout` (bool, optional, default=True): passed to
            <autofig.figure.Figure.draw> if `show` or `save`.  Whether to draw
            with the `tight_layout` option.
        * `draw_title` (bool, optional, default=True): passed to
            <autofig.figure.Figure.draw> if `show` or `save`.  Whether to draw
            the title on the matplotlib axes.
        * `subplot_grid` (None or tuple, optional, default=None): passed to
            <autofig.figure.Figure.draw> if `show` or `save`.  Override the
            subplot locations.
        * `show` (bool, optional, default=False): whether to immediately
            draw and show the resulting matplotlib figure.  If True,
            <autofig.figure.Figure.draw> will be called.
        * `save` (False or string, optional, default=False): the filename
            to save the resulting matplotlib figure, or False to not save.
            If not False, <autofig.figure.Figure.draw> will be called.
        * `**kwargs`: additional keyword arguments are passed on to
            <autofig.call.FillBetween.__init__> to initialize the new
            <autofig.call.FillBetween>.
        """

        tight_layout = kwargs.pop('tight_layout', True)
        draw_sidebars = kwargs.pop('draw_sidebars', True)
        draw_title = kwargs.pop('draw_title', True)
        subplot_grid = kwargs.pop('subplot_grid', None)

        show = kwargs.pop('show', False)
        save = kwargs.pop('save', False)

        call = _call.FillBetween(*args, **kwargs)
        self.add_call(call)

        if show or save:
            self.reset_draw()
            return self.draw(tight_layout=tight_layout,
                             draw_sidebars=draw_sidebars,
                             draw_title=draw_title,
                             subplot_grid=subplot_grid,
                             show=show, save=save)

    @property
    def meshes(self):
        """
        Access all children <autofig.call.Mesh>es of the <autofig.figure.Figure>.

        See also:

        * <autofig.figure.Figure.calls>
        * <autofig.figure.Figure.plots>
        * <autofig.figure.Figure.fill_betweens>

        Returns
        -------------
        * <autofig.call.MeshGroup> of all <autofig.call.Mesh> objects.
        """
        calls = [c for c in self._calls if isinstance(c, _call.Mesh)]
        return _call.MeshGroup(calls)

    def mesh(self, *args, **kwargs):
        """
        Add a new <autofig.call.Mesh> to the <autofig.figure.Figure>.

        See also:

        * <autofig.call.Mesh.__init__>

        Arguments
        ----------
        * `*args`: all positional arguments are passed on to
            <autofig.call.Mesh.__init__> to initialize the new
            <autofig.call.Mesh>.
        * `tight_layout` (bool, optional, default=True): passed to
            <autofig.figure.Figure.draw> if `show` or `save`.  Whether to draw
            with the `tight_layout` option.
        * `draw_title` (bool, optional, default=True): passed to
            <autofig.figure.Figure.draw> if `show` or `save`.  Whether to draw
            the title on the matplotlib axes.
        * `subplot_grid` (None or tuple, optional, default=None): passed to
            <autofig.figure.Figure.draw> if `show` or `save`.  Override the
            subplot locations.
        * `show` (bool, optional, default=False): whether to immediately
            draw and show the resulting matplotlib figure.  If True,
            <autofig.figure.Figure.draw> will be called.
        * `save` (False or string, optional, default=False): the filename
            to save the resulting matplotlib figure, or False to not save.
            If not False, <autofig.figure.Figure.draw> will be called.
        * `**kwargs`: additional keyword arguments are passed on to
            <autofig.call.Mesh.__init__> to initialize the new
            <autofig.call.Plot>.
        """

        tight_layout = kwargs.pop('tight_layout', True)
        draw_sidebars = kwargs.pop('draw_sidebars', True)
        draw_title = kwargs.pop('draw_title', True)
        subplot_grid = kwargs.pop('subplot_grid', None)
        show = kwargs.pop('show', False)
        save = kwargs.pop('save', False)

        call = _call.Mesh(*args, **kwargs)
        self.add_call(call)
        if show or save:

            self.reset_draw()
            return self.draw(tight_layout=tight_layout,
                             draw_sidebars=draw_sidebars,
                             draw_title=draw_title,
                             subplot_grid=None,
                             show=show, save=save)

    # def show(self):
    #     plt.show()

    def reset_draw(self):
        """
        Clear the underlying matplotlib figure object.
        """
        # TODO: figure options like figsize, etc

        fig = self._get_backend_object(naxes=len(self.axes))
        fig.clf()

    def draw(self, fig=None, i=None, calls=None,
             tight_layout=True,
             draw_sidebars=True,
             draw_title=True,
             subplot_grid=None,
             show=False, save=False,
             save_afig=False,
             in_animation=False):
        """
        Draw the contents of the <autofig.figure.Figure> to a matplotlib figure
        object.

        See also:

        * <autofig.figure.Figure.animate>
        * <autofig.draw>
        * <autofig.axes.Axes.draw>
        * <autofig.call.Plot.draw>
        * <autofig.call.Mesh.draw>

        Arguments
        ------------
        * `fig` (matplotlib figure or None, optional, default=None): matplotlib
            figure instances to use during drawing.
        * `i` (float or None, optional, default=None): passed on to
            <autofig.axes.Axes.draw> for all <autofig.axes.Axes> in
            <autofig.figure.Figure.axes>.
        * `calls` (list of <autofig.call.Call> objects or None, optional, default=None):
            passed on to <autofig.axes.Axes.draw> for all <autofig.axes.Axes> in
            <autofig.figure.Figure.axes>.
        * `tight_layout` (bool, optional, default=True): whether to call
            `fig.tight_layout` after positioning all subplots, but before
            drawing any applicable sidebars.
        * `draw_sidebars` (bool, optional, default=True): whether to draw
            any applicable sidebars.
        * `draw_title` (bool, optional, default=True): passed on to
            <autofig.axes.Axes.draw> for all <autofig.axes.Axes> in
            <autofig.figure.Figure.axes>.  Whether to draw the title on the
            matplotlib axes.
        * `subplot_grid` (None or tuple, optional, default=None): Override the
            subplot locations.  Passed on to <autofig.axes.Axes.append_subplot>
            for each <autofig.axes.Axes> in <autofig.figure.Figure.axes>.
        * `show` (bool, optional, default=False): whether to immediately
            draw and show the resulting matplotlib figure.
        * `save` (False or string, optional, default=False): the filename
            to save the resulting matplotlib figure, or False to not save.
        * `save_afig` (False or string, optional, default=False): the filename
            to save the autofig object, along with the options for this
            draw call.  See also <autofig.figure.Figure.save>.
        * `in_animation` (bool, optional, default=False): whether the current
            call to `draw` is a single frame in an animation.  Usually this
            should not be changed by the user.  See <autofig.figure.Figure.animate>
            for creating animations.

        Returns
        ----------
        * ([matplotlib Figure](https://matplotlib.org/api/_as_gen/matplotlib.figure.Figure.html#matplotlib.figure.Figure)): the matplotlib figure object.
        """

        if save_afig:
            render = {'render': 'draw'}
            render['i'] = i
            render['tight_layout'] = tight_layout
            render['draw_sidebars'] = draw_sidebars
            render['draw_title'] = draw_title
            render['subplot_grid'] = subplot_grid

            self.save(save_afig, renders=[render])

        fig = self._get_backend_object(fig, naxes=len(self.axes))
        callbacks._connect_to_autofig(self, fig)
        callbacks._connect_to_autofig(self, fig.canvas)

        if calls is None:
            # then we need to reset the backend figure.  This is especially
            # important when passing draw(i=something)
            fig.clf()

        for axesi in self.axes:
            if axesi._backend_object not in fig.axes:
                # then axes doesn't have a subplot yet.  Adding one will also
                # shift the location of all axes already drawn/created.
                ax = axesi.append_subplot(fig=fig, subplot_grid=subplot_grid)
                # if axesi._backend_object already existed (but maybe on a
                # different figure) it will be reset on the draw call below.
            else:
                # then this axes already has a subplot on the figure, so we'll
                # allow it to default to that instance
                ax = None

            axesi.draw(ax=ax, i=i, calls=calls,
                       draw_sidebars=False,
                       draw_title=draw_title,
                       show=False, save=False, in_animation=in_animation)

            self._backend_artists += axesi._get_backend_artists()

        # must call tight_layout BEFORE adding any sidebars
        if tight_layout:
            fig.tight_layout()

        if draw_sidebars:
            for axesi in self.axes:
                axesi.draw_sidebars(i=i)

        if save:
            fig.savefig(save)

        if show:
            # TODO: allow top-level option for whether to block or not?
            if not common._inline:
                plt.show()  # <-- blocking
                # fig.show()  #<-- not blocking

        return fig

    def animate(self, fig=None, i=None,
                tight_layout=False,
                draw_sidebars=True,
                draw_title=True,
                subplot_grid=None,
                interval=100,
                animate_callback=None,
                show=False, save=False, save_kwargs={},
                save_afig=False):
        """
        Draw the contents of the <autofig.figure.Figure> to a matplotlib animation.

        See also:

        * <autofig.figure.Figure.plot>

        Arguments
        ------------
        * `fig` (matplotlib figure or None, optional, default=None): matplotlib
            figure instances to use during drawing.
        * `i` (list/array, **required**): iterable values for `i`.  Each item
            in the list/array will become a single frame in the resulting
            animation.
        * `tight_layout` (bool, optional, default=True): whether to call
            `fig.tight_layout` after positioning all subplots, but before
            drawing any applicable sidebars.
        * `draw_sidebars` (bool, optional, default=True): whether to draw
            any applicable sidebars.
        * `draw_title` (bool, optional, default=True): passed on to
            <autofig.axes.Axes.draw> for all <autofig.axes.Axes> in
            <autofig.figure.Figure.axes>.  Whether to draw the title on the
            matplotlib axes.
        * `subplot_grid` (None or tuple, optional, default=None): Override the
            subplot locations.  Passed on to <autofig.axes.Axes.append_subplot>
            for each <autofig.axes.Axes> in <autofig.figure.Figure.axes>.
        * `interval` (int, optional, default=100): time in ms between each
            frame in the animation.
        * `animate_callback` (callable, optional, default=None): Function which
            takes the matplotlib figure object and will be called at each frame
            within the animation.
        * `show` (bool, optional, default=False): whether to immediately
            draw and show the resulting matplotlib animation.
        * `save` (False or string, optional, default=False): the filename
            to save the resulting matplotlib animation, or False to not save.
        * `save_kwargs` (dict, optional, default={}): dictionary of keyword
            arguments to be passed on to [anim.save](https://matplotlib.org/api/_as_gen/matplotlib.animation.FuncAnimation.html#matplotlib.animation.FuncAnimation.save)
        * `save_afig` (False or string, optional, default=False): the filename
            to save the autofig object, along with the options for this
            animate call.  See also <autofig.figure.Figure.save>.

        Returns
        ----------
        * ([matplotlib FuncAnimation](https://matplotlib.org/api/_as_gen/matplotlib.animation.FuncAnimation.html#matplotlib-animation-funcanimation)): the matplotlib animation object.
        """

        if save_afig:
            render = {'render': 'animate'}
            render['i'] = i
            render['tight_layout'] = tight_layout
            render['draw_sidebars'] = draw_sidebars
            render['draw_title'] = draw_title
            render['subplot_grid'] = subplot_grid
            render['interval'] = interval

            self.save(save_afig, renders=[render])

        if tight_layout:
            print("WARNING: tight_layout with fixed limits may cause jittering in the animation")

        if i is None:
            # TODO: can we get i from the underlying Axes/Calls?
            raise NotImplementedError("must pass a list/array for i")

        if not hasattr(i, '__iter__'):
            raise ValueError("i must be iterable for animations")


        # time interval in ms between each frame
        interval = int(interval)

        # TODO: use blitting (probably only if no mesh calls)
        # https://matplotlib.org/3.1.0/api/animation_api.html#funcanimation
        blit = False

        ao = _mpl_animate.Animation(self,
                                    tight_layout=tight_layout,
                                    draw_sidebars=draw_sidebars,
                                    draw_title=draw_title,
                                    subplot_grid=subplot_grid,
                                    animate_callback=animate_callback)

        anim = animation.FuncAnimation(ao.mplfig, ao, fargs=(),\
                init_func=ao.anim_init, frames=i, interval=interval,\
                blit=blit)

        if save:
            try:
                anim.save(save, **save_kwargs)
            except ValueError as err:
                if str(err)=='I/O operation on closed file':
                    raise ValueError("saving animation failed (with message: {}). Try passing a valid option to 'write' via save_kwargs.  For example: save_kwargs={{'writer': 'imagemagick'}}".format(str(err)))
                else:
                    traceback.print_exc()

        if show:
            # TODO: allow top-level option for whether to block or not?
            if not common._inline:
                plt.show()  # <-- blocking
                # fig.show()  #<-- not blocking

        return anim




from phoebe.parameters import *
from phoebe import conf, __version__

### NOTE: if creating new parameters, add to the _forbidden_labels list in parameters.py

def settings(**kwargs):
    """
    Create a <phoebe.parameters.ParameterSet> for bundle-level settings.

    Generally, this will automatically be added to a newly initialized
    <phoebe.frontend.bundle.Bundle>

    Arguments
    ----------
    * `dict_filter` (dictionary, optional, default={}): filter to use when using
        dictionary access in the bundle.
    * `dict_set_all` (bool, optional, default=False): whether to set all values
        for dictionary access that returns more than 1 results.
    * `run_checks_compute` (list or string, optional, default='*'): Compute
        options to use when calling run_checks/run_checks_compute or within
        interactive checks.
    * `run_checks_solver` (list or string, optional, default='*'): Solver
        options to use when calling run_checks/run_checks_solver or within
        interactive checks.
    * `run_checks_solution` (list or string, optional, default='*'): Solutions
        to use when calling run_checks/run_checks_solution or within
        interactive checks.
    * `run_checks_figure` (list or string, optional, default='*'): Figures
        to use when calling run_checks/run_checks_figure or within
        interactive checks.
    * `auto_add_figure` (bool, optional, default=False): Whether to automatically
        add figure parameters when a dataset is added with a new dataset type,
        or a solution is added.
    * `auto_remove_figure` (bool, optional, default=False): Whether to
        automatically remove figure parameters when the referenced
        dataset/solution are removed.
    * `web_client` (bool, optional, default=False): Whether to default to using
        the web-client over a locally installed desktop-client when opening the
        UI from the desktop client.
    * `web_client_url` (string, optional, default='ui.phoebe-project.org'):
        Default location of web-client.  Will only be used if web_client is True.

    Returns
    --------
    * (<phoebe.parameters.ParameterSet>): ParameterSet of all newly created
        <phoebe.parameters.Parameter> objects.
    """

    params = []

    params += [StringParameter(qualifier='phoebe_version', value=kwargs.get('phoebe_version', __version__), advanced=True, readonly=True, description='Version of PHOEBE')]
    params += [DictParameter(qualifier='dict_filter', value=kwargs.get('dict_filter', {}), advanced=True, description='Filters to use when using dictionary access')]
    params += [BoolParameter(qualifier='dict_set_all', value=kwargs.get('dict_set_all', False), advanced=True, description='Whether to set all values for dictionary access that returns more than 1 result')]

    params += [SelectParameter(qualifier='run_checks_compute', value=kwargs.get('run_checks_compute', '*'), choices=[], advanced=False, description='Compute options to use when calling run_checks/run_checks_compute or within interactive checks.')]
    params += [SelectParameter(qualifier='run_checks_solver', value=kwargs.get('run_checks_solver', '*'), choices=[], advanced=False, description='Solver options to use when calling run_checks/run_checks_solver or within interactive checks.')]
    params += [SelectParameter(qualifier='run_checks_solution', value=kwargs.get('run_checks_solution', []), choices=[], advanced=False, description='Solutions to use when calling run_checks/run_checks_solution or within interactive checks.')]
    params += [SelectParameter(qualifier='run_checks_figure', value=kwargs.get('run_checks_figure', []), choices=[], advanced=False, description='Figures to use when calling run_checks/run_checks_figure or within interactive checks.')]

    params += [BoolParameter(qualifier='auto_add_figure', value=kwargs.get('auto_add_figure', True), description='Whether to automatically add figure parameters when a dataset is added with a new dataset type, or a solution is added.')]
    params += [BoolParameter(qualifier='auto_remove_figure', value=kwargs.get('auto_remove_figure', True), description='Whether to automatically remove figure parameters when the referenced dataset/solution are removed.')]

    params += [BoolParameter(qualifier='web_client', value=kwargs.get('web_client', False), advanced=True, description='Whether to default to using the web-client over a locally installed desktop-client when opening the UI from the desktop client.')]
    params += [StringParameter(qualifier='web_client_url', value=kwargs.get('web_client_url', 'ui.phoebe-project.org'), advanced=True, description='Default location of web-client.  Will only be used if web_client is True.')]

    return ParameterSet(params)



from phoebe.parameters import *
#from phoebe.parameters.unit_choices import unit_choices as _unit_choices
#import numpy as np

#from .common import _use_autofig, _mplcmaps, _mplcolors, _mplmarkers, _mpllinestyles, MPLPropCycler, _label_units_lims, _figure_style_sources, _figure_style_nosources, _figure_uncover_highlight_animate

def distribution_collection(b, **kwargs):
    params = []

    params += [ChoiceParameter(qualifier='distribution_set', value=kwargs.get('distribution_set', 'manual'), choices=['manual'], description='Set of distributions to include in the plot.  Use \'manual\' to explicitly select distributions and combining logic')]
    params += [SelectParameter(visible_if='distribution_set:manual', qualifier='distributions', value=kwargs.get('distributions', []), choices=[''], description='Distributions to include in the plot')]
    params += [ChoiceParameter(visible_if='distribution_set:manual', qualifier='combine', value=kwargs.get('combine', 'and'), choices=['and', 'or', 'first'], description='How to combine multiple distributions for the same parameter.')]
    params += [BoolParameter(visible_if='distribution_set:manual', qualifier='include_constrained', value=kwargs.get('include_constrained', False), description='Whether to include constrained parameters')]
    params += [BoolParameter(visible_if='distribution_set:manual', qualifier='to_uniforms', value=kwargs.get('to_uniforms', False), description='Wheter to convert all distribution to univariate uniforms')]
    params += [IntParameter(visible_if='distribution_set:manual,to_uniforms:True', qualifier='to_uniforms_sigma', value=kwargs.get('to_uniforms_sigma', 1), description='Sigma to adopt for non-uniform distributions')]
    params += [BoolParameter(visible_if='distribution_set:manual,to_uniforms:False', qualifier='to_univariates', value=kwargs.get('to_univariates', False), description='Whether to convert any multivariate distribution to univaritates before adding to the collection')]

    params += [IntParameter(qualifier='titles_sigma', value=kwargs.get('titles_sigma', 1), limits=[0, None], description='Sigma to adopt for uncertainties in axes titles (or 0 to disable).')]

    return ParameterSet(params)
